# routines to compute static charge-charge correlators
from numba import jit
from .. import algebra
import numpy as np


def chargechi(h,i=0,j=0):
    """Compute the charge correlator for a Hamiltonian"""
    if h.has_eh: 
        print("Hamiltonians with eh not implemented")
        raise # not implemented
    if not h.has_spin:
        return single_chargechi(h,i=i,j=j)
    else:
        up = single_chargechi(h,i=2*i,j=2*j) 
        dn = single_chargechi(h,i=2*i+1,j=2*j+1) 
        ud = single_chargechi(h,i=2*i+1,j=2*j) 
        du = single_chargechi(h,i=2*i,j=2*j+1) 
        return up + dn + ud + du # return both contributions



def szchi(h,i=0,j=0):
    """Compute the charge correlator for a Hamiltonian"""
    if h.has_eh:
        print("Hamiltonians with eh not implemented")
        raise # not implemented
    if not h.has_spin:
        return single_chargechi(h,i=i,j=j)
    else:
        up = single_chargechi(h,i=2*i,j=2*j)
        dn = single_chargechi(h,i=2*i+1,j=2*j+1)
        ud = single_chargechi(h,i=2*i,j=2*j+1) 
        du = single_chargechi(h,i=2*i+1,j=2*j) 
        return (up + dn - ud -du)/4. # return both contributions


def sxchi(H,**kwargs):
    """Compute magnetic response in the x axis"""
    from .. import rotate_spin
    vector = [0.,1.,0.] # rotation vector
    H = H.copy() # make a copy
    phi = 1.0/2.
    rotate_spin.hamiltonian_spin_rotation(H,vector=vector,angle=phi)
    return szchi(H,**kwargs)




def sychi(H,**kwargs):
    """Compute magnetic response in the y axis"""
    from .. import rotate_spin
    vector = [1.,0.,0.] # rotation vector
    H = H.copy() # make a copy
    phi = 1.0/2.
    rotate_spin.hamiltonian_spin_rotation(H,vector=vector,angle=phi)
    return szchi(H,**kwargs)





def single_chargechi(h,i=0,j=0,temp=1e-7):
    """Compute charge response function for a single orbital"""
    if h.dimensionality!=0: raise
    hk = h.get_hk_gen() # get generator
    m = hk(0) # get Hamiltonian
    esh,ws = algebra.eigh(m) # diagonalize
    ws = np.transpose(ws) # transpose wavefunctions
    if i<0: raise # sanity check
    if j<0: raise # sanity check
    return elementchi(ws,esh,ws,esh,i,j,temp)



@jit(nopython=True)
def elementchi(ws1,es1,ws2,es2,ii,jj,T):
    """Compute the response function"""
    out = 0j # initialize
    n = len(ws1) # number of wavefunctions
    for i in range(n): # first loop over states
      oi = (-np.tanh(es1[i]/T) + 1.0)/2. # first occupation
      for j in range(n): # second loop over states
          oj = (-np.tanh(es1[j]/T) + 1.0)/2. # first occupation
          fac = np.conjugate(ws1[i][ii])*ws2[j][ii] # add the factor
          fac *= ws1[i][jj]*np.conjugate(ws2[j][jj]) # add the factor
          # probably this should be written better
          fac *= oi*(1.- oj) # occupation factor
#          fac *= abs(oi - oj)/2. # occupation factor
          out = out + fac
    return out






import numpy as np

def fermidirac(es,temp=0.):
    """Return a Fermi Dirac distribution at a certain temperature"""
    if temp==0.: return (1.0 - np.sign(es))/2. # zero temperature
    else: return 1./(1. + np.exp(es/temp))


def dFDdT(es,temp=0.):
    """Return a Fermi Dirac distribution at a certain temperature"""
    if temp==0.: raise # not implemented
    else: return 1./temp*np.exp(es/temp)/(1. + np.exp(es/temp))**2



# routines to work with wannier hamiltonians
from __future__ import print_function
import numpy as np
from . import hamiltonians
from . import geometry
from scipy.sparse import csc_matrix as csc
from scipy.sparse import bmat
from scipy.sparse import coo_matrix
from . import multicell
from . import angular
import scipy.linalg as lg
from copy import deepcopy
import os
from . import filesystem as fs


def convert(inname="wannier.win",name="wannier"):
  """Read a wannier.win file, and create a folder with
  all the necessary files that characterize the Hamiltonian"""
  fs.rmdir(name) # remove folder
  fs.mkdir(name) # create folder





def read_geometry(input_file="wannier.win"):
  """Reads the geometry of the wannier calculation"""
  ll = read_between("begin unit_cell_cart","end unit_cell_cart",input_file)
  a1 = ll[1].split()
  a2 = ll[2].split()
  a3 = ll[3].split()
  # read the unit vectors
  a1 = np.array([float(a1[0]),float(a1[1]),float(a1[2])]) # first vector
  a2 = np.array([float(a2[0]),float(a2[1]),float(a2[2])]) # second vector
  a3 = np.array([float(a3[0]),float(a3[1]),float(a3[2])]) # second vector
  g = geometry.Geometry()
  g.dimensionality = 2
  g.has_spin = False
  g.has_sublattice = False
  g.a1 = a1  # store vector
  g.a2 = a2  # store vector
  g.a3 = a3  # store vector
  # read the coordinates 
  ll = read_between("begin projections","end projections",input_file)
  rs = [] # empty list for positions
  g.atoms_have_names = True # Atoms have names
  g.atoms_names = [] # initalize
  for l in ll:
    name = l.split(":")[0] # get name of the atom
    r = get_positions(name,input_file) # get positins of the atoms
    g.atoms_names += [name]*len(r) # append this names
    for i in range(len(r)): # to real coordinates
      r[i] = r[i][0]*a1 + r[i][1]*a2 + r[i][2]*a3
    rs += r # store positions
  g.r = np.array(rs) # store in the class
  g.r2xyz() # store also in xyz atributes
  g.get_fractional() # fractional coordinates
  return g


def get_positions(atom,input_file):
  """Get positions of certain orbitals"""
  ll = read_between("begin atoms_frac","end atoms_frac",input_file)
  rs = [] # empty list
  for l in ll: # loop over lines
    l = l.split()
    name = l[0]
    if atom==name: # found atom
      r = np.array([float(l[1]),float(l[2]),float(l[3])]) # position
      rs.append(r) # add to the list
  return rs # return positions


def get_projected_atom_names(input_file):
  """Get the names of all the atoms who have projections"""
  lorb = read_between("begin projections","end projections",input_file)
  names = []
  for l in lorb:
    name = l.split(":")[0]
    if name not in names:
      names.append(name) # add this name
  return names


def get_orbitals(specie,input_file="wannier.win"):
  """ Get the names of the orbitals of a cetain atom"""
  lorb = read_between("begin projections","end projections",input_file)
  orbs = []
  for l in lorb: # loop over orbitals
    sname = l.split(":")[0] # name of the atom
    oname = l.split(":")[1] # name of the orbital
    oname = oname.split()[0] # remove the \n
    if specie==sname: orbs.append(oname)
  return orbs

def get_all_orbitals(input_file="wannier.win"):
  """Get the name of all the orbitals"""
  atoms = get_projected_atom_names(input_file)
  nout = []
  for a in atoms:
    for io in get_orbitals(a,input_file=input_file): # loop over orbitals
      nat = get_atoms_specie(a,input_file=input_file) # number of atoms
      nout += [a+str(i)+"," + io for i in range(nat)]
  return nout



def get_atoms_specie(specie,input_file="wannier.win"):
  """ Get number of atoms o a certain specie"""
  ll = read_between("begin atoms_frac","end atoms_frac",input_file)
  nat = 0
  for l in ll:
    name = l.split()[0] # name of the atom
    if name==specie: nat +=1 
  return nat # return umber of atoms






def get_index_orbital(specie,atom,orbital,input_file="wannier.win"):
  """Get the index of a certain orbital"""
  ll = read_between("begin atoms_frac","end atoms_frac",input_file)
  lorb = read_between("begin projections","end projections",input_file)
  iorb = 0 # index in the matrix
  print(specie)
  for l in lorb: # loop over orbitals
    sname = l.split(":")[0] # name of the atom
    oname = l.split(":")[1] # name of the orbital
    oname = oname.split()[0] # remove the \n
    ifound = 0 # number of atoms found
    for la in ll: # loop over atoms in the structure
      rname = la.split()[0] # name of the atom in the structure
      if rname==sname: 
        iorb += 1 # increase de counter of the hamiltonian index 
        ifound += 1 # increase the counter in the atoms
      # if the desired orbital and atom has been reached
      if specie==sname and atom==ifound and orbital==oname:
        return iorb
  raise # error if this point is reached



def get_indexes(input_file="wannier.win"):
  """Returns a list with the indexes of the different orbitals for
  each atom """
  ll = read_between("begin atoms_frac","end atoms_frac",input_file)
  lorb = read_between("begin projections","end projections",input_file)
  # first get all the different names
  names = get_projected_atom_names(input_file) # get names of the atoms
  dat = dict() # create a diccionary
  for name in names: # loop over different atoms
    out = [] # empty list with the different indexes and names
    ind = 0 # initialize index of the hamiltonian
    for l in ll: # loop over atoms in the structure
      out_atom = [] # empty list for the next atom
      l = l.split() # split the line
      label = l[0]  # get label of the atom
      if label==name: # if same name as input
        for lo in lorb: # go to the orbitals list and look for this atom
          name_o = lo.split(":")[0] # get name of the atom in the orbital list
          if name_o==name: # if an orbital of this atom has been found 
            orb = lo.split(":")[1] # name of the orbital
            orb = orb.split()[0] # name of the orbital
            out_atom.append((ind,orb)) # append tuple with index and norm
        out.append([out_atom]) # append this atom to the final 
    dat[name] = out # dd this atom to the diccionary
  return dat  # return the diccionary








def read_between(a,b,input_file):
  ll = open(input_file).readlines()
  start = False # found the klist
  out = []
  for (i,l) in zip(range(len(ll)),ll):
    if b in l: break # end of klist
    if start: # sotre line
      out.append(l)
    if a in l: start = True # found beginning 
  return out # return output lines





def read_hamiltonian(input_file="hr_truncated.dat",is_real=False):
  """Reads an output hamiltonian from wannier"""
  mt = np.genfromtxt(input_file) # get file
  m = mt.transpose() # transpose matrix
  # read the hamiltonian matrices
  class Hopping: pass # create empty class
  tlist = []
  def get_t(i,j,k):
    norb = np.max([np.max(np.abs(m[3])),np.max(np.abs(m[4]))])
    norb = int(norb)
    mo = np.matrix(np.zeros((norb,norb),dtype=np.complex_))  
    for l in mt: # look into the file
      if i==int(l[0]) and j==int(l[1]) and k==int(l[2]):
        if is_real:
          mo[int(l[3])-1,int(l[4])-1] = l[5] # store element
        else:
          mo[int(l[3])-1,int(l[4])-1] = l[5] + 1j*l[6] # store element
    return mo # return the matrix
#  for i in range(-nmax,nmax):
#    for j in range(-nmax,nmax):
#      for k in range(-nmax,nmax):
#        t = Hopping() # create hopping
#        t.dir = [i,j,k] # direction
#        t.m = get_t(i,j,k) # read the matrix
#        tlist.append(t) # store hopping
  # the previous is not used yet...
  g = geometry.kagome_lattice() # create geometry
  h = g.get_hamiltonian() # build hamiltonian
  h.intra = get_t(0,0,0)
  h.tx = get_t(1,0,0)
  h.ty = get_t(0,1,0)
  h.txy = get_t(1,1,0)
  h.txmy = get_t(1,-1,0)
  h.has_spin = False  # if not spin polarized
  h.geometry = read_geometry() # read the geometry of the system
  if len(h.geometry.r)!=len(h.intra): 
    print("Dimensions do not match",len(g.r),len(h.intra))
    print(h.geometry.r)
    raise # error if dimensions dont match
  return h



def read_multicell_hamiltonian(input_file="hr_truncated.dat",
                                ncells=None,win_file="wannier.win",
                                dim=2,skip_win=False,path=None):
  """Reads an output hamiltonian from wannier"""
  if path is not None: 
      inipath = os.getcwd() # current path
      os.chdir(path) # go there
  mt = np.genfromtxt(input_file) # get file
  m = mt.transpose() # transpose matrix
  if ncells is None: # use all the hoppings
    nmax = int(np.max([np.max(m[i])for i in range(3)]))
    ncells = [nmax,nmax,nmax]
  # read the hamiltonian matrices
  class Hopping: pass # create empty class
  tlist = []
  def get_t(i,j,k):
    norb = np.max([np.max(np.abs(m[3])),np.max(np.abs(m[4]))])
    norb = int(norb)
    mo = np.matrix(np.zeros((norb,norb),dtype=np.complex_))  
    found = False
    for l in mt: # look into the file
      if i==int(l[0]) and j==int(l[1]) and k==int(l[2]):
        mo[int(l[3])-1,int(l[4])-1] = l[5] + 1j*l[6] # store element
        found  = True # matrix has been found
    if found:  return mo # return the matrix
    else: return None # return nothing if not found
  for i in range(-ncells[0],ncells[0]+1):
    for j in range(-ncells[1],ncells[1]+1):
      for k in range(-ncells[2],ncells[2]+1):
        if (i,j,k)==(0,0,0): continue # skip intracell
        matrix = get_t(i,j,k) # read the matrix
        if matrix is None: continue
        else: # store hopping
          t = Hopping() # create hopping
          t.dir = [i,j,k] # direction
          t.m = get_t(i,j,k) # read the matrix
          tlist.append(t) # store hopping
  # the previous is not used yet...
  g = geometry.kagome_lattice() # create geometry
  h = g.get_hamiltonian() # build hamiltonian
  h.is_multicell = True
  if not skip_win: # do not skip wannier.win
    h.orbitals = get_all_orbitals(input_file=win_file)
  h.hopping = tlist # list of hoppings
  h.has_spin = False  # if not spin polarized
  if not skip_win: # do not skip readin wannier.win
    h.geometry = read_geometry(input_file=win_file) # read the geometry of the system
    h.geometry.center() # center the geometry
  h.intra = get_t(0,0,0)
  if not skip_win: # do not skip reading wannier.win
    if len(h.geometry.r)!=len(h.intra): 
      print("Dimensions do not match",len(g.r),len(h.intra))
      print(h.geometry.r)
  #  raise # error if dimensions dont match
  h.dimensionality = dim 
  if path is not None: 
      os.chdir(inipath) # go back
      h.wannierpath = path # store
  else: h.wannierpath = None
  # now lets add the SOC method
  def get_soc(self,name,soc):
      if not self.has_spin: raise # only for spinful
      self.intra = self.intra + generate_soc(name,soc,path=self.wannierpath) 
  import types
  h.get_soc = types.MethodType(get_soc,h) # add the method
  return h







def get_klist(input_file="wannier.win",nkpoints=500):
  """ Get the klist for bands calculation"""
  ll = read_between("begin kpoint_path","end kpoint_path",input_file)
  kp = [] # empty vertex
  for l in ll: 
    l2 = l.split() # split the numbers
    kp.append([[float(l2[1]),float(l2[2]),float(l2[3])],[float(l2[5]),float(l2[6]),float(l2[7])]])
  klist = [] # empty klist
  nk = int(nkpoints/len(kp)) # number of kpoints
  for (k1,k2) in kp: # loop over pairs
    k1 = np.array(k1)
    k2 = np.array(k2)
    dk = (k2-k1)/nk # create dk
    for i in range(nk): # loop over ks
      klist.append(k1+i*dk)
  return klist


def get_num_wannier(input_file):
  """Get the number of wannier orbitals"""
  ll = read_between("begin atoms_frac","end atoms_frac",input_file)
  lorb = read_between("begin projections","end projections",input_file)
  norb = 0
  for o in lorb: # loop over orbitals
    name = o.split(":")[0] # name of the orbital
    for l in ll:
      rname = l.split()[0] # name of the atom
      if name==rname: norb += 1
  return norb # return number of orbitals




def write_wannier_names(input_file="wannier.win",output_file="ORBITALS.OUT"):
  """Get the number of wannier orbitals"""
  ll = read_between("begin atoms_frac","end atoms_frac",input_file)
  lorb = read_between("begin projections","end projections",input_file)
  norb = 0
  fo = open(output_file,"w") # open file to write
  fo.write("# Name of the atom,    index of this type of atom, name of the orbital\n")
  for o in lorb: # loop over orbitals
    iatom = 1 # initialize
    name = o.split(":")[0] # name of the orbital
    oname = o.split(":")[1] # name of the orbital
    oname = oname.split()[0]
    for l in ll:
      rname = l.split()[0] # name of the atom
      if name==rname: # if this atom found
        fo.write(name + "              " +str(iatom)+"               "+oname+"\n") # new line
        iatom += 1 # increase atom counter
  fo.close()
  print("Orbitals written to",output_file)









def get_num_atoms(specie,input_file):
  """Get the number of wannier orbitals"""
  ll = read_between("begin atoms_frac","end atoms_frac",input_file)
  nat = 0 # initialize
  for l in ll:
    name = l.split()[0] # name of the atom
    if name==specie: nat += 1
  return nat # return number of orbitals






def generate_soc(specie,value,input_file="wannier.win",nsuper=1,path=None):
  """Add SOC to a hamiltonian based on wannier.win"""
  if path is not None: 
      inipath = os.getcwd() # current path
      os.chdir(path) # go there
  o = open(".soc.status","w")
  iat = 1 # atom counter
  orbnames = names_soc_orbitals(specie) # get which are the orbitals
  ls = soc_l((len(orbnames)-1)/2) # get the SOC matrix
  norb = get_num_wannier(input_file) # number of wannier orbitals
  m = np.matrix([[0.0j for i in range(norb*2)] for j in range(norb*2)]) # matrix
  nat = get_num_atoms(specie,input_file) # number of atoms of this specie
  for iat in range(nat):
   # try:
#      fo.write("Attempting "+specie+"  "+str(iat+1)+"\n")
      for i in range(len(orbnames)): # loop over one index
        orbi = orbnames[i] 
        for j in range(len(orbnames)): # loop over other index
          orbj = orbnames[j]
          ii = get_index_orbital(specie,iat+1,orbi)  # index in wannier
          jj = get_index_orbital(specie,iat+1,orbj) # index in wannier
#          fo.write(str(ii)+"   "+str(jj)+"\n")
          ii += -1 # python starts in 0
          jj += -1 # python starts in 0
          m[2*ii,2*jj] = ls[2*i,2*j] # store the soc coupling
          m[2*ii+1,2*jj] = ls[2*i+1,2*j] # store the soc coupling
          m[2*ii,2*jj+1] = ls[2*i,2*j+1] # store the soc coupling
          m[2*ii+1,2*jj+1] = ls[2*i+1,2*j+1] # store the soc coupling
   #   return
   # except: break
#  fo.close()
  n = nsuper**2 # supercell
  mo = [[None for i in range(n)] for j in range(n)]
  for i in range(n): mo[i][i] = csc(m) # diagonal elements
  mo = bmat(mo).todense() # dense matrix
  if path is not None: 
      os.chdir(inipath) # go there
      print(np.max(np.abs(mo)))
  return np.matrix(mo*value) # return matrix
#  for name in atoms: # loop over atoms



def names_soc_orbitals(specie):
  return angular.names_soc_orbitals(specie)




def ylm2xyz_l2(has_spin=True):
  """Return the matrix that converts the cartesian into spherical harmonics"""
  m = angular.ylm2xyz_l2() # return change of bassi matrix
  if has_spin:
    from .increase_hilbert import spinful
    m = spinful(m) # with spin degree of freedom
  return m



def ylm2xyz_l1(has_spin=True):
  """Return the matrix that converts the cartesian into spherical harmonics"""
  m = angular.ylm2xyz_l1() # return change of bassi matrix
  if has_spin:
    from .increase_hilbert import spinful
    m = spinful(m) # with spin degree of freedom
  return m







def soc_l(l):
  """Calculate the spin orbit coupling in a basis of spherical harmonics"""
  l = int(l)
  nm = 2*l + 1 # number of components
  zero = np.matrix([[0.0j for i in range(2*nm)] for j in range(2*nm)]) 
  # initialize matrices
  lz = zero.copy()
  lx = zero.copy()
  ly = zero.copy()
  lm = zero.copy()
  lp = zero.copy()
  # create l+ and l- and lz
  for m in range(-l,l): # loop over m components
    val = np.sqrt((l-m)*(l+m+1)) # value of the cupling
    im = m + l
    lp[2*(im+1),2*im] = val # up channel
    lp[2*(im+1)+1,2*im+1] = val # down channel
  for m in range(-l,l+1): # loop over m components
    im = m + l
    lz[2*im,2*im] = m # value of lz, up channel
    lz[2*im+1,2*im+1] = m # value of lz, down channel
  lm = lp.H # adjoint
  lx = (lp + lm) /2.
  ly = -1j*(lp - lm) /2.
  # create spin matrices
  sz = zero.copy()
  sx = zero.copy()
  sy = zero.copy()
  for m in range(-l,l+1): # loop over m components
    im = m + l
    sx[2*im,2*im+1] = 1.0 
    sx[2*im+1,2*im] = 1.0 
    sy[2*im,2*im+1] = -1j 
    sy[2*im+1,2*im] = 1j 
    sz[2*im,2*im] = 1.
    sz[2*im+1,2*im+1] = -1.
  # check that the matrix is fine
  sx = sx/2.
  sy = sy/2.
  sz = sz/2.
  if True:
    comm_zero(sx,lx)
    comm_zero(sx,ly)
    comm_zero(sx,lz)
    comm_zero(sy,ly)
    comm_zero(sz,lz)
    comm_angular(sx,sy,sz)
    comm_angular(sy,sz,sx)
    comm_angular(sz,sx,sy)
    comm_angular(lx,ly,lz)
    comm_angular(ly,lz,lx)
    comm_angular(lz,lx,ly)
  ls = lx*sx + ly*sy + lz*sz  # SOC matrix
  import scipy.linalg as lg
  from scipy.sparse import csc_matrix as csc
  if l==2: R = ylm2xyz_l2() # get change of basis matrix
  if l==1: R = ylm2xyz_l1() # get change of basis matrix
  ls = R.H * ls * R # change to cartesian orbitals
  return ls # return the matrix



def comm_angular(x,y,z):
  from scipy.sparse import csc_matrix as csc
  xy = x*y - y*x
  xy = xy - 1j*z
  if np.abs(np.max(xy))>0.001:
    raise



def comm_zero(x,y):
  xy = x*y - y*x
  if np.abs(np.max(xy))>0.01:
    raise



def symmetrize_atoms(h,specie,input_file="wannier.win"):
  """Symmetrizes a certain atom"""
  orbs = get_orbitals(specie,input_file=input_file) # read the orbitals
  nat = get_atoms_specie(specie,input_file=input_file) # number of atoms
  if h.has_spin: raise
  for iorb in orbs: # loop over orbitals
    avg = 0.
    for iat in range(nat): # loop over atoms 
      i = get_index_orbital(specie,iat+1,iorb) - 1 
      ons = h.intra[i,i] # add to the average
      avg += ons # add to the average
    avg = avg/nat # average value
    for iat in range(nat): # loop over atoms
      i = get_index_orbital(specie,iat+1,iorb) - 1
      h.intra[i,i] = avg # substitute by average


def get_hoppings(h,specie1,specie2,input_file="wannier.win"):
  """Get hoppings between two atoms"""
  orbs1 = get_orbitals(specie1,input_file=input_file) # read the orbitals
  orbs2 = get_orbitals(specie2,input_file=input_file) # read the orbitals
  nat1 = get_atoms_specie(specie1,input_file=input_file) # number of atoms
  nat2 = get_atoms_specie(specie2,input_file=input_file) # number of atoms
  mats = [h.intra,h.tx,h.ty,h.txy,h.txmy]
  for iorb1 in orbs1: # loop over orbitals
    for iorb2 in orbs2: # loop over orbitals
      for iat1 in range(nat1): # loop over atoms 
        for iat2 in range(nat2): # loop over atoms 
          i = get_index_orbital(specie1,iat1+1,iorb1) - 1 
          j = get_index_orbital(specie2,iat2+1,iorb1) - 1 
          for m in mats:
            ons = m[i,j].real # add to the average



def get_atomic_projection(specie,input_file="wannier.win",has_spin=False):
  """Get the matrix that projects onto a certain atom"""
  orbs = get_orbitals(specie,input_file=input_file) # read the orbitals
  nat = get_atoms_specie(specie,input_file=input_file) # number of atoms
  norb = get_num_wannier(input_file) # number of wannier orbitals
  proj = np.matrix([[0.0j for i in range(norb)] for j in range(norb)])
  for iat in range(nat): # loop over atoms of this type
    for iorb in orbs: # loop over orbitals
      i = get_index_orbital(specie,iat+1,iorb) - 1 
      proj[i,i] = 1.0 # non vanishing
  if has_spin:
    from .increase_hilbert import spinful
    return spinful(proj)
  else: return proj





def read_supercell_hamiltonian(input_file="hr_truncated.dat",is_real=False,nsuper=1):
  """Reads an output hamiltonian for a supercell from wannier"""
  mt = np.genfromtxt(input_file) # get file
  m = mt.transpose() # transpose matrix
  # read the hamiltonian matrices
  class Hopping: pass # create empty class
  tlist = []
  def get_t(i,j,k):
    norb = int(np.max([np.max(np.abs(m[3])),np.max(np.abs(m[4]))]))
    mo = np.matrix(np.zeros((norb,norb),dtype=np.complex_))  
    for l in mt: # look into the file
      if i==int(l[0]) and j==int(l[1]) and k==int(l[2]):
        if is_real:
          mo[int(l[3])-1,int(l[4])-1] = l[5] # store element
        else:
          mo[int(l[3])-1,int(l[4])-1] = l[5] + 1j*l[6] # store element
    return mo # return the matrix
  # this function will be called in a loop
  g = geometry.kagome_lattice() # create geometry
  h = g.get_hamiltonian() # build hamiltonian
  h.has_spin = False
  nstot = nsuper**2
  intra = [[None for i in range(nstot)] for j in range(nstot)]
  tx = [[None for i in range(nstot)] for j in range(nstot)]
  ty = [[None for i in range(nstot)] for j in range(nstot)]
  txy = [[None for i in range(nstot)] for j in range(nstot)]
  txmy = [[None for i in range(nstot)] for j in range(nstot)]
  from scipy.sparse import csc_matrix as csc
  vecs = []
  # create the identifacion vectors
  inds = []
  acu = 0
  try: # read different supercells
    nsuperx = nsuper[0]
    nsupery = nsuper[1]
    nsuperz = nsuper[2]
  except: # read different supercells
    nsuperx = nsuper
    nsupery = nsuper
    nsuperz = nsuper
  for i in range(nsuperx): # loop over first replica
    for j in range(nsupery): # loop over second replica
      vecs.append(np.array([i,j])) # append vector
      inds.append(acu)
      acu += 1 # add one to the accumulator
  for i in inds: # loop over first vector
    for j in inds:  # loop over second vector
      v1 = vecs[i] # position of i esim cell
      v2 = vecs[j] # position of j esim cell
      dv = v2 - v1 # difference in vector
      # get the different block elements
      intra[i][j] = csc(get_t(dv[0],dv[1],0))
      tx[i][j] = csc(get_t(dv[0]+nsuper,dv[1],0))
      ty[i][j] = csc(get_t(dv[0],dv[1]+nsuper,0))
      txy[i][j] = csc(get_t(dv[0]+nsuper,dv[1]+nsuper,0))
      txmy[i][j] = csc(get_t(dv[0]+nsuper,dv[1]-nsuper,0))
  h.intra = bmat(intra).todense()
  h.tx = bmat(tx).todense()
  h.ty = bmat(ty).todense()
  h.txy = bmat(txy).todense()
  h.txmy = bmat(txmy).todense()
  h.geometry = read_geometry() # read the geometry of the system
  if nsuper>1:
    h.geometry = h.geometry.supercell(nsuper) # create supercell
  if len(h.geometry.r)!=len(h.intra): 
    print("Dimensions do not match",len(g.r),len(h.intra))
    print(h.geometry.r)
 #   raise # error if dimensions dont match
  # names of the orbitals
  h.orbitals = get_all_orbitals()*nsuper**2

  return h




def rotate90(h):
  """ Rotate by 90 degrees th unit cell"""
  h.geometry.x, h.geometry.y = h.geometry.y, h.geometry.x
  h.geometry.xyz2r()
  a1 = h.geometry.a1
  a2 = h.geometry.a2
  h.geometry.a1, h.geometry.a2 = np.array([a2[1],a2[0],0.]),  np.array([a1[1],a1[0],0.])  # geometry
  h.tx , h.ty = h.ty, h.tx
  h.txy , h.tmy = h.txy, h.txmy.H
  return h




def save(output_file="wannier.xml",hamiltonian=None):
  """Saves a wannier hamiltonian in xml format"""
  import xml.etree.cElementTree as ET
  strf = "{:10.8f}".format  # float format
  if hamiltonian is None: # no hamiltonian provided
    hamiltonian = read_multicell_hamiltonian() # read hamiltonian
  else:
    hamiltonian = multicell.turn_multicell(hamiltonian)
  root = ET.Element("wannier") # root element
  try: orbs = hamiltonian.orbitals # name of the orbitals
  except: orbs = ["X,Y" for i in range(hamiltonian.intra.shape[0])]
  geo = hamiltonian.geometry # geometry of the crystal
  #### start write the orbitals ###
  orbxml = ET.SubElement(root, "orbitals") # subelement orbitals
  fo = open("orbitals.wan","w") # human readable
  ffrac = open("fractional_coordinates.wan","w") # human readable
  freal = open("real_coordinates.wan","w") # human readable
  fo.write("# index, name,       position\n")
  ffrac.write("# position\n")
  freal.write("# position\n")
  for i in range(len(geo.r)):
    orbital = ET.SubElement(orbxml, "orbital") # subelement orbital
    ET.SubElement(orbital, "index").text = str(i)
    ET.SubElement(orbital, "atom_name").text = orbs[i].split(",")[0]
    ET.SubElement(orbital, "orbital_name").text = orbs[i].split(",")[1]
    ET.SubElement(orbital, "x").text = strf(geo.x[i])
    ET.SubElement(orbital, "y").text = strf(geo.y[i])
    ET.SubElement(orbital, "z").text = strf(geo.z[i])
    ET.SubElement(orbital, "frac_x").text = strf(geo.frac_x[i])
    ET.SubElement(orbital, "frac_y").text = strf(geo.frac_y[i])
    ET.SubElement(orbital, "frac_z").text = strf(geo.frac_z[i])
    # human readable
    fo.write("    "+str(i+1)+"    "+orbs[i]+"    ")
    fo.write(strf(geo.x[i])+"    "+strf(geo.y[i])+"   "+strf(geo.z[i])+"  ")
    freal.write(strf(geo.x[i])+"    "+strf(geo.y[i])+"   "+strf(geo.z[i])+"\n")
    fo.write(strf(geo.frac_x[i])+"    "+strf(geo.frac_y[i])+"   "+strf(geo.frac_z[i])+"\n")
    ffrac.write(strf(geo.frac_x[i])+"      "+strf(geo.frac_y[i])+"      "+strf(geo.frac_z[i])+"\n")
  #### end write the orbitals ###
  fo.close()
  ffrac.close()
  freal.close()
  #### start write the geometry ###

  geoxml = ET.SubElement(root, "geometry") # subelement orbitals
  aname = ["a1","a2","a3"] # name of unit vectors
  aobj = [geo.a1,geo.a2,geo.a3]
  fg = open("geometry.wan","w") # geometry file
  fg.write("# unit cell vectors\n")
  for (an,ao) in zip(aname,aobj): # loop over vectors
    a1xml = ET.SubElement(geoxml, an) # subelement orbital
    ET.SubElement(a1xml, "x").text = strf(ao[0])
    ET.SubElement(a1xml, "y").text = strf(ao[1])
    ET.SubElement(a1xml, "z").text = strf(ao[2])
    fg.write(strf(ao[0])+"  "+strf(ao[1])+"  "+strf(ao[2])+"\n")
  fg.close()
  #### end write the geometry ###

  #### start write the hamiltonian ###
  
  ham = ET.SubElement(root, "hamiltonian") # subelement hamiltonian
  fo = open("hamiltonian.wan","w") # human readable
  fo.write("# nx, ny, nz, i, j, real, imaginary\n")
  # intracell object (workaround)
  intraobj = deepcopy(hamiltonian.hopping[0])
  intraobj.dir = [0,0,0] # store direction
  intraobj.m = hamiltonian.intra # store matrix
  for hop in [intraobj]+hamiltonian.hopping:
    d = hop.dir # direction
    m = hop.m # hopping matrix
    m = coo_matrix(m) # to coo matrix
    ii = m.row # row index
    jj = m.col # column idex
    data = m.data # data
    for (i,j,mij) in zip(ii,jj,data): # loop over elements
      dist = geo.r[i] - (geo.r[j]+d[0]*geo.a1+d[1]*geo.a2+d[2]*geo.a3)
      dist = np.sqrt(dist.dot(dist))
      element = ET.SubElement(ham, "hopping") # subelement hamiltonian
      ET.SubElement(element, "nx").text = str(d[0]) # element
      ET.SubElement(element, "ny").text = str(d[1]) # element
      ET.SubElement(element, "nz").text = str(d[2]) # element
      ET.SubElement(element, "i").text = str(i+1) # element
      ET.SubElement(element, "j").text = str(j+1) # element
      ET.SubElement(element, "iname").text = orbs[i] # element
      ET.SubElement(element, "jname").text = orbs[j] # element
      ET.SubElement(element, "real_amplitude").text = strf(mij.real) # element
      ET.SubElement(element, "imag_amplitude").text = strf(mij.imag) # element
      ET.SubElement(element, "distance").text = strf(dist) # element

      # human readable
      fo.write("   "+str(d[0])+"   "+str(d[1])+"   "+str(d[2])+"   ") 
      fo.write(str(i+1)+"    "+str(j+1)+"    ")
      fo.write(strf(mij.real)+"    "+strf(mij.imag)+"\n")
  fo.close() # close hamiltonian file
  #### end write the hamiltonian ###


  #### save everything ###
  tree = ET.ElementTree(root) # create whole tree
  tree.write("wannier.xml") # write tree
  import xml.dom.minidom
  xml = xml.dom.minidom.parse("wannier.xml") 
  open("wannier.xml","w").write(xml.toprettyxml())





def save_readable(hamiltonian,output_file="hamiltonian.wan"):
  """Saves a wannier hamiltonian in xml format"""
  strf = "{:10.8f}".format  # float format
  hamiltonian = multicell.turn_multicell(hamiltonian)
  #### start write the hamiltonian ###
  fo = open(output_file,"w") # human readable
  fo.write("# nx, ny, nz, i, j, real, imaginary\n")
  # intracell object (workaround)
  class ObjHop: pass
  intraobj = ObjHop()
  intraobj.dir = [0,0,0] # store direction
  intraobj.m = hamiltonian.intra # store matrix
  for hop in [intraobj]+hamiltonian.hopping:
    d = hop.dir # direction
    m = hop.m # hopping matrix
    m = coo_matrix(m) # to coo matrix
    ii = m.row # row index
    jj = m.col # column idex
    data = m.data # data
    for (i,j,mij) in zip(ii,jj,data): # loop over elements
      # human readable
      fo.write("   "+str(d[0])+"   "+str(d[1])+"   "+str(d[2])+"   ") 
      fo.write(str(i+1)+"    "+str(j+1)+"    ")
      fo.write(strf(mij.real)+"    "+strf(mij.imag)+"\n")
  fo.close() # close hamiltonian file
  #### end write the hamiltonian ###










import numpy as np




class Tensor3():
    """Class for a three dimensional tensor"""
    def __init__(self,row,col,pla,data,shape=None):
        if shape is None: raise
        self.row = row
        self.col = col
        self.pla = pla
        self.data = data
        self.shape = shape
    def __mul__(self,a):
        """Multiply by a vector, assuming that it will use
        the third component, and return an array"""
        if a.shape[0]!= self.shape[2]: raise # inconsistent dimensions
        out = np.zeros((self.shape[0],self.shape[1]),dtype=np.complex_)
        for ii in range(len(self.row)):
            i = self.row[ii]
            j = self.col[ii]
            k = self.pla[ii]
            d = self.data[ii]
            out[i,j] += a[k]*d # add contribution
        return out # return matrix
    def copy(self):
        from copy import deepcopy
        return deepcopy(self)
    def __add__(self,a):
        """Sum two tensors"""
        if type(a)!=Tensor3: raise # not implemented
        if a.shape[0]!=self.shape[0]: raise
        if a.shape[1]!=self.shape[1]: raise
        if a.shape[2]!=self.shape[2]: raise
        out = self.copy() # copy object
        out.row = np.concatenate([self.row,a.row])
        out.col = np.concatenate([self.col,a.col])
        out.pla = np.concatenate([self.pla,a.pla])
        out.data = np.concatenate([self.data,a.data])
        return out # return object



class TensorN():
    """Define an N dimensional sparse tensor"""
    def __init__(self,index,data,shape=None):
        self.index = index
        self.data = data
        self.shape = shape







import numpy as np
from scipy.sparse import issparse,coo_matrix,csc_matrix


def spin_mixing_part(m):
    """Return the spin mixing part of a matrix"""
    m = coo_matrix(m) # convert
    n = m.shape[0]//2
    d1,r1,c1 = [],[],[]
    d0,r0,c0 = m.data,m.row,m.col
    for i in range(len(d0)):
        if r0[i]%2==0 and c0[i]%2==1:
            d1.append(d0[i])
            r1.append(r0[i]//2)
            c1.append(c0[i]//2)
    mo = csc_matrix((d1,(r1,c1)),shape=(n,n),dtype=np.complex_)
    if issparse(m): return np.array(mo.todense())
    else: return mo



def full2profile(h,profile,check=True):
  """Resums a certain profile to show only the spatial dependence"""
  n = len(profile)
  if check:
    if len(profile)!=h.intra.shape[0]: raise # inconsistency
  if h.has_spin == False and h.has_eh==False: out = np.array(profile)
  elif h.has_spin == True and h.has_eh==False:
    out = np.array([profile[2*i]+profile[2*i+1] for i in range(n//2)])
  elif h.has_spin == False and h.has_eh==True:
    out = np.array([profile[2*i]+profile[2*i+1] for i in range(n//2)])
  elif h.has_spin == True and h.has_eh==True:
    out = np.array([profile[4*i]+profile[4*i+1]+profile[4*i+2]+profile[4*i+3] for i in range(n//4)])
  else: raise # unknown
  if check:
    if len(out)!=len(h.geometry.r): raise # mistmach in the dimensions
  return out






from __future__ import print_function
import scipy.linalg as lg
import numpy as np

# library to get function that estimate the importance of each point
# in reciprocal space

def kmesh_density(h,nk=20,delta=0.01):
  """Function to compute a k-mesh density districution"""
  if h.dimensionality !=2: raise
  hkgen = h.get_hk_gen() # get generator
  out = [] # empty list
  xout = []
  yout = []
  for x in np.linspace(-0.1,1.1,nk,endpoint=True):
    for y in np.linspace(-0.1,1.1,nk,endpoint=True):
      hk = hkgen([x,y])
      es = lg.eigvalsh(hk) # eigenvalues
      e = np.min(es[es>0.]) - np.max(es[es<0.]) # gap
      out.append(e) # append gap
      xout.append(x)
      yout.append(y)
  # now create the function that interpolates
  import interpolation
  f = interpolation.interpolator2d(xout,yout,out) # interpolator function
  # now define a function that returns a probability density
  def fout(k):
    return delta/(delta**2 + f(k[0],k[1])) # output
  return fout # return function



def weighted_mesh(h,n=10000,delta=0.1):
  """Return a set of random k-points and their relative weight
  using the gap as a probability measure"""
  kx = np.random.random(n) # kx coordinate
  ky = np.random.random(n) # ky coordinate








import numpy as np

class Mixing():
    def __init__(self,n=6):
        """Create the mixing object"""
        self.n = n
        self.errors = np.zeros(n) # empty array
        self.full = False
        self.i = 0 # number of stored
    def get_mix(self,d):
        """Get the new mixing"""
        self.errors = np.concatenate([self.errors[1:self.n],[d]]) # store
        if self.full: # filled mixing
            m = self.errors[self.n-1]/self.errors[self.n-2] # decrease
            if abs(m-1.0)<1e-2: return 0.01 # slow mixing
            elif m>1.: return 0.5 # increasing, normal mixing
            else: return 1.-m
        else:
            self.i += 1 # increase counter
            if self.i == (self.n-1): self.full = True
            return 0.5





import numpy as np



def label2k(g,kl):
    """Given a kpoint label, return the kpoint"""
    if kl=="G": return [0.,0.,0.]
    elif kl=="M": return [0.5,0.,0.]
    elif kl=="K":
        from .locate import target_moduli
        from .mapping import unitary
        vm = np.sqrt(3) # moduli
        k = g.k2K(target_moduli(unitary(g.b1),unitary(g.b2),vm))
        return k*2/(3.*np.sqrt(3))
    elif kl=="K'": return -label2k(g,"K")
    elif kl=="M1": return [.5,.0,.0]
    elif kl=="M2": return [.0,.5,.0]
    elif kl=="M3": return [.5,.5,.0]
    elif kl=="X": return [.5,.0,.0]
    elif kl=="Y": return [.0,.5,.0]
    else: 
        print("Unrecognized kpoint",kl)
        raise






# compute vacuum expectation values

import numpy as np
from operators import Operator

def get_dm_vev(H,A,**kwargs):
    """Compute a vacuum expectation value of two operators"""
    if H.dimensionality != 0:
        print("Only implemented for 0d Hamiltonians")
        raise
    dm = H.get_density_matrix() # return the DM, as a matrix
    A = Operator(A) # convert to operator
    return np.trace(A@dm) # return the expectation value




import numpy as np
import scipy.sparse as sp
from scipy.sparse import bmat
from .reorder import reorder
from .. import algebra

def hopping2deltaud(H,T):
    """Given a hopping object, add a delta ud with the exact same
    form to the Hamiltonian. This can be fast for very large systems"""
    from ..superconductivity import reorder
    H.turn_nambu() # turn the Nambu spinor
    H.turn_multicell() # multicell mode
    T = T.copy() # make a dummy copy
    T.remove_spin() # remove the spin degree of freedom
    T.turn_multicell() # multicell mode
    n = len(H.geometry.r) # number of sites
    def t2h(m): # transform a single hopping to deltaud
      m = algebra.todense(m)
      pout = [[None for i in range(n)] for j in range(n)] # initialize
      iden = sp.identity(2) # identity
      for i in range(n): pout[i][i] = iden*0.
      for i in range(n): # loop over sites
        for j in range(n): # loop over sites
            if np.abs(m[i,j])<1e-6: continue # if the hopping is zero, skip
            else: pout[i][j] = m[i,j]*iden # get this pairing
      diag = sp.identity(2*n)*0. # zero matrix
      pout = bmat(pout) # convert to block matrix
      mout = [[diag,pout],[None,diag]] # output matrix
      mout = bmat(mout) # return full matrix
      return reorder(mout) # reorder the entries properly
    T.intra = t2h(T.intra) # intracell hopping
    for hop in T.hopping:
        hop.m = t2h(hop.m) # convert to pairing by hand
    T.has_eh = True # add eh by hand
    T.has_spin = True # add spin by hand
    # now the complementary
    T2 = T.copy()
    T2.intra = algebra.dagger(T2.intra)
    for hop in T2.hopping:
        hop.m = algebra.dagger(hop.m) # the other
        hop.dir = tuple(-np.array(hop.dir)) # the other
#    return T + T2 # summ both terms via brute force
    return H + T + T2 # summ both terms via brute force




import numpy as np
from scipy.sparse import coo_matrix,bmat
from .rotate_spin import sx,sy,sz
from . import neighbor

def add_rashba(self,c):
    """
    Add rashba to a Hamiltonian
    """
    from .operators import isnumber
    if isnumber(c):
        if c==0.0: return
    if not self.has_spin: # no spin degree of freedom
        self.turn_spinful() # spinful hamiltonian
    def rashba(*args,**kwargs):
        return self.spinful2full(rashba_matrix(*args,**kwargs))
    g = self.geometry
    is_sparse = self.is_sparse # saprse Hamiltonian
    self.intra = self.intra + rashba(g.r,c=c,is_sparse=is_sparse)
    if self.dimensionality==0: return
    if self.is_multicell: # multicell hamiltonians
      for i in range(len(self.hopping)): # loop over hoppings
        d = self.hopping[i].dir # direction
        Rd = g.a1*d[0] + g.a2*d[1] + g.a3*d[2]
        r2 = [ir + Rd for ir in g.r] # new coordinates
        self.hopping[i].m = self.hopping[i].m + rashba(g.r,r2=r2,c=c,is_sparse=is_sparse)
    else: # conventional Hamiltonians
      if g.dimensionality==1:  # one dimensional
        r2 = [ir + g.a1 for ir in g.r]
        self.inter = self.inter + rashba(g.r,r2=r2,c=c,is_sparse=is_sparse)
      elif g.dimensionality==2:  # two dimensional
        r2 = [ir + g.a1 for ir in g.r]
        self.tx = self.tx + rashba(g.r,r2=r2,c=c,is_sparse=is_sparse)
        r2 = [ir + g.a2 for ir in g.r]
        self.ty = self.ty + rashba(g.r,r2=r2,c=c,is_sparse=is_sparse)
        r2 = [ir + g.a1+g.a2 for ir in g.r]
        self.txy = self.txy + rashba(g.r,r2=r2,c=c,is_sparse=is_sparse)
        r2 = [ir + g.a1-g.a2 for ir in g.r]
        self.txmy = self.txmy + rashba(g.r,r2=r2,c=c,is_sparse=is_sparse)
      else: raise








def rashba_matrix(r1,r2=None,c=0.,d=[0.,0.,1.],is_sparse=False):
  """
  Add Rashba coupling, returns a spin polarized matrix
  This will assume only Rashba between first neighbors
  """
  zero = coo_matrix([[0.,0.],[0.,0.]])
  if r2 is None:
    r2 = r1
  nat = len(r1) # number of atoms
  m = [[None for i in range(nat)] for j in range(nat)] # create matrix
  for i in range(nat): m[i][i] = zero.copy() # initilize
  neighs = neighbor.connections(r1,r2) # neighbor connections
  for i in range(nat): # loop over first atoms
#    for j in range(nat):  # loop over second atoms
    for j in neighs[i]:  # loop over second atoms
      rij = r2[j] - r1[i]   # x component
      dx,dy,dz = rij[0],rij[1],rij[2]  # different components
      rxs = [dy*sz-dz*sy,dz*sx-dx*sz,dx*sy-dy*sx]  # cross product
      ms = 1j*(d[0]*rxs[0] + d[1]*rxs[1] + d[2]*rxs[2]) # E dot r times s
      s = 0.0*ms
      if 0.9<(rij.dot(rij))<1.1: # if nearest neighbor
        if callable(c): s = ms*c((r1[i]+r2[j])/2.) # function
        else: s = ms*c # multiply
      m[i][j] = s # rashba term
  if not is_sparse: m = bmat(m).todense()  # to normal matrix
  else: m = bmat(m) # sparse matrix
  return m








def clean_hamiltonian(h):
  """Set a Hamiltonian to zero"""
  h.intra *= 0.0
  if h.is_multicell:
    for t in h.hopping: t.m *= 0.0
  else:
    if h.dimensionality==0: pass
    elif h.dimensionality==1: h.inter *= 0.0
    elif h.dimensionality==2: 
      h.tx *= 0.0
      h.ty *= 0.0
      h.txy *= 0.0
      h.txmy *= 0.0
    elif h.dimensionality==3: raise




import numpy as np
from . import geometry


remove_duplicated = geometry.remove_duplicated_positions

### Routines dealing with Kekule order ###

def kekule_positions(r):
    """
    Returns the positions defining a Kekule ordering
    """
    cs = hexagon_centers(r,r) # return the centers
    cs = remove_duplicated(cs) # remove duplicated
    zs = np.unique(np.round(cs[:,2],5)) # unique z positions
    cso = [] # output list
    for z in zs:
        csi = cs[(np.abs(cs[:,2]-z)<1e-4),:] # this layer
        csi = retain(csi,d=3.0) # retain only centers that are at distance 3
        cso = cso + np.array(csi).tolist() # store
    return np.array(cso) # return array



def kekule_center(r):
    """
    Returns a function that given two positions, returns the Kekule center
    """
    cs = kekule_positions(r) # returns centers of the kekule
    def f(r1,r2): # function that returs the kekule center
        dr = r1-r2 ; dr=dr.dot(dr) 
        if not 0.99<dr<1.01: return None # too far
        for c in cs: # loop over centers
            # ensure that they are in the "hexagon"
            dr1 = c-r1 ; dr1 = dr1.dot(dr1) # distance to r1
            dr2 = c-r2 ; dr2 = dr2.dot(dr2) # distance to r2
            if 0.99<dr1<1.01 and 0.99<dr2<1.01: # if clse to center
              return c # return the center
        return None # no center found
    return f # return function






def kekule_function(r,t=1.):
    """
    Returns a function that will compute Kekule hoppings
    """
    cs = kekule_positions(r) # get the centers with all the positions
    ## Define a function to only have hoppings in the hexagon
#    if callable(t): fun = t # t is a function
#    else: # assume t is a number
    if callable(t): tfun = t # t is a function
    else: tfun = lambda r: t # t is a number
    def fun(r1,r2):
        dr = r1-r2 ; dr = dr.dot(dr) # distance 1
        if 0.99<dr<1.01: return tfun((r1+r2)/2.) # first neighbor
        else: return 0.0 # nothing
    def f(r1,r2):
        for c in cs: # loop over centers
            dr = r1-r2 ; dr=dr.dot(dr) 
            if not 0.99<dr<1.01: continue
            # ensure that they are in the "hexagon"
            dr1 = c-r1 ; dr1 = dr1.dot(dr1) # distance to r1
            dr2 = c-r2 ; dr2 = dr2.dot(dr2) # distance to r2
            if 0.99<dr1<1.01 and 0.99<dr2<1.01: # if clse to center
              return fun(r1,r2) # return the hopping
        return 0.0 # no hopping
    # now define the function
    def fm(rs1,rs2):
      m = np.zeros((len(rs1),len(rs2)),dtype=np.complex_) # initialize matrix
      for i in range(len(rs1)): # loop
        for j in range(len(rs2)): # loop
            m[i,j] = f(rs1[i],rs2[j]) # get kekule coupling
      return m # return the Kekule matrix
    return fm # return the function


def kekule_matrix(r1,r2=None,**kwargs):
    """
    Return a Kekule matrix for positions r, assuming
    they are from a honeycomb-like lattice
    """
    if r2 is None: r2 = r1
    f = kekule_function(r1,**kwargs)
    return f(r1,r2)


def hexagon_centers(r1,r2):
    """
    Return the centers of an hexagon
    """
    out = []
    for ri in r1: # loop
        for rj in r2: # loop
            dr = ri-rj
            dr = dr.dot(dr) # distance
            if 3.9<dr<4.1: # center of an hexagon
                out.append((ri+rj)/2.) # store the center
    return out # return list with centers


def r_in_rs(r,rs):
    """
    Check that a position is not stored
    """
    for ri in rs:
        dr = ri-r ; dr = dr.dot(dr)
        if dr<0.01: return True
    return False



def retain(r,d=3.0):
    """
    Retain only sites that are at a distance d
    """
    i = np.random.randint(len(r))
    out = [r[0]] # take first one
    def iterate(out): # do one iteration
      out0 = [r for r in out] # initialize
      for rj in out: # loop over stored
        for ri in r: # loop
            dr = (ri-rj) ; dr = dr.dot(dr) # distance
            if d*d-0.1<dr<d*d+0.1: # if desired distance
                # now check that this one has not been stored already
                if not r_in_rs(ri,out0): # not stored yet
                  out0.append(ri) # store position
      return out0
#    np.savetxt("R.OUT",np.matrix(r))
#    exit()
    while True:
#    for i in range(10):
        out1 = iterate(out) # do one iteration
        out1 = remove_duplicated(out1) # remove duplicated atoms
        if len(out1)==len(out): break
        out = [r for r in out1] # redefine
#    np.savetxt("R.OUT",np.matrix(out)) # write in file
#    exit()
    return out # return desired positions




def chiral_kekule(g,t1=0.0,t2=0.0,hermitian=True):
    """Add a chiral kekule hopping, input is a geometry"""
    fkc = kekule_center(g.r) # function for kekule center
    dr0 = g.r[0] - g.r[1] # one NN vector
    dz = np.exp(1j*2.*np.pi/3.) # second vector
    z0 = dr0[0] + 1j*dr0[1] # complex NN vector
    if not 0.99<dr0.dot(dr0)<1.01: raise # this has to be more robust
    # function to check first kind of bond
    def kind1(dr):
        z = dr[0] + 1j*dr[1] # complex position
        for iz in [z0,z0*dz,z0*dz*dz]: # loop
            if abs(z*np.conjugate(iz)-1.0)<0.01: return True
        return False
    # function to check if going clockwise
    clockwise = lambda r1,r2: np.cross(r1,r2)[2]>0. # if clockwise
    tz = dz # complex hopping
    tz = 1.0
    def fun(r1,r2):
        dr = r1-r2 # distance between sites
        if not 0.99<dr.dot(dr)<1.01: return 0. # not first neighbors
        rk = fkc(r1,r2) # Kekule center
        if clockwise(rk-r1,r2-r1): # clockwise hopping
          if kind1(dr): return t1 # first kind of bond
          else: return t2
        else:
          if hermitian: # conventional Hamiltonian
            if kind1(dr): return np.conjugate(t2) # first kind of bond
            else: return np.conjugate(t1)
          else: return 0.0
    return fun 








import numpy as np

def kdos(self,kpath=None,energies=None,
           write=True,**kwargs):
    """Compute momentum-resolved spectral function"""
    def fun(k,e):
        if self.dimensionality==2: # 2D heterostructure
            HT1 = self.generate(k) # generate heterostructure
            return HT1.get_coupled_central_dos(energy=e,**kwargs)
        else: raise # not implemented
    if kpath is None: kpath = np.linspace(0.,1.,40)
    if energies is None: energies = np.linspace(-1.0,1.,40)
    energies = np.array(energies)
    from ..parallel import pcall
    kout,eout,dout = [],[],[]
    ds = pcall(lambda k: [fun(k,e) for e in energies],kpath) # call in parallel
    for (k,d) in zip(kpath,ds): # loop over kpoints
        kout = np.concatenate([kout,energies*0.+k]) # store kpoint
        eout = np.concatenate([eout,energies]) # store energies
        dout = np.concatenate([dout,d]) # store DOS
    if write:
        np.savetxt("KDOS.OUT",np.array([kout,eout,dout]).T)
    return (kout,eout,dout)




import numpy as np
from .. import algebra,operators

# workaround for non hermitian Hamiltonians


def get_bands_nd(h,kpath=None,operator=None,num_bands=None,
                    callback=None,central_energy=0.0,nk=400,
                    ewindow=None,eigmode="complex",
                    output_file="BANDS.OUT",write=True,
                    silent=True):
    """
    Get an n-dimensional bandstructure
    """
    if num_bands is not None:
      if num_bands>(h.intra.shape[0]-1): num_bands=None
    if operator is not None: operator = h.get_operator(operator)
    if num_bands is None: # all the bands
      if operator is not None: 
        def diagf(m): # diagonalization routine
            return algebra.eig(m) # all eigenvals and eigenfuncs
      else: 
        def diagf(m): # diagonalization routine
            return algebra.eigvals(m) # all eigenvals and eigenfuncs
    else: # using arpack
      h = h.copy()
      h.turn_sparse() # sparse Hamiltonian
      def diagf(m):
        eig,eigvec = slg.eigs(m,k=num_bands,which="LM",sigma=central_energy,
                                    tol=arpack_tol,maxiter=arpack_maxiter)
        if operator is None: return eig
        else: return (eig,eigvec)
    # open file and get generator
    hkgen = h.get_hk_gen() # generator hamiltonian
    kpath = h.geometry.get_kpath(kpath,nk=nk) # generate kpath
    def getek(k):
      """Compute this k-point"""
      out = [] # output list
      hk = hkgen(kpath[k]) # get hamiltonian
      if operator is None: # just compute the energies
        es = diagf(hk)
        es = np.sort(es) # sort energies
        for e in es:  # loop over energies
            out.append([k,e])
        if callback is not None: callback(k,es) # call the function
      else:
        es,ws = diagf(hk)
        ws = ws.transpose() # transpose eigenvectors
        def evaluate(w,k,A): # evaluate the operator
            if type(A)==operators.Operator:
                waw = A.braket(w,k=kpath[k]).real
            elif callable(A):  
              try: waw = A(w,k=kpath[k]) # call the operator
              except: 
                print("Check out the k optional argument in operator")
                waw = A(w) # call the operator
            else: waw = braket_wAw(w,A).real # calculate expectation value
            return waw # return the result
        for (e,w) in zip(es,ws):  # loop over waves
          if callable(ewindow):
              if not ewindow(e): continue # skip iteration
          if isinstance(operator, (list,)): # input is a list
              waws = [evaluate(w,k,A) for A in operator]
          else: waws = [evaluate(w,k,operator)]
          oi = [k,e] # create list
          for waw in waws:  oi.append(waw) # add this one
          out.append(oi) # store
        # callback function in each iteration
        if callback is not None: callback(k,es,ws) # call the function
      return out # return string
    if True:
      esk = [] # empty list
      for k in range(len(kpath)): # loop over kpoints
        ek = getek(k)
        esk += ek # add lists
    esk = np.array(esk).T
    if eigmode=="complex": pass # full eigenvalue
    elif eigmode=="real": esk[1] = esk[1].real # real part of eigenvalue
    elif eigmode=="imag": esk[1] = esk[1].imag # imag part of eigenvalue
    else: raise
    return esk





# woorkaround for non Hermitian Hamiltonians

from ..dos import get_dos_general

def get_dos(self,mode="",**kwargs):
    # overwrite the method to use ED
    return get_dos_general(self,mode="ED",**kwargs)




# different types of disorder for a heterostructure

import numpy as np
import random

def anderson(hetero,w=0.2,write=False):
  """Adds anderson disorder, asumes spinpol calculation"""
  hc = hetero.central_intra # this is a nxn tridiagonal list of matrices
  nb = len(hc)  # number of blocks
  norb = len(hc[0][0]) # number of orbitals with spin degree of freedom
  if hetero.has_spin:
    norb = norb/2 # without spin degree
  if write: fa = open("ANDERSON_DISORDER.OUT","w")
  for ib in range(nb): # loop over blocks
    for iorb in range(norb): # loop over orbitals
      wi = (random.random()-0.5)*w  # stregth of the disorder
      if hetero.has_spin:
        hc[ib][ib][2*iorb,2*iorb] += wi  # up channel
        hc[ib][ib][2*iorb+1,2*iorb+1] += wi  # same for down channel
      else:
        hc[ib][ib][iorb,iorb] += wi  # spinless channel
      if write: # write in file
        fa.write(str(ib)+"   "+str(iorb)+"   "+str(wi)+"\n") # save the quantity
  if write: fa.close()  # close file


def magnetic_z(hetero,w=0.2):
  """Adds off-plane magnetic disorder, asumes spinpol calculation"""
  hc = hetero.central_intra # this is a nxn tridiagonal list of matrices
  nb = len(hc)  # number of blocks
  norb = len(hc[0][0]) # number of orbitals with spin degree of freedom
  norb = norb/2 # without spin degree
  fa = open("MZ_DISORDER.OUT","w")
  for ib in range(nb): # loop over blocks
    for iorb in range(norb): # loop over orbitals
      wi = (random.random()-0.5)*w  # stregth of the disorder
      hc[ib][ib][2*iorb,2*iorb] += wi
      hc[ib][ib][2*iorb+1,2*iorb+1] += -wi
      fa.write(str(ib)+"   "+str(iorb)+"   "+str(wi)+"\n") # save the quantity
  fa.close()



def magnetic_x(hetero,w=0.2):
  """Adds in-plane magnetic disorder, asumes spinpol calculation"""
  hc = hetero.central_intra # this is a nxn tridiagonal list of matrices
  nb = len(hc)  # number of blocks
  norb = len(hc[0][0]) # number of orbitals with spin degree of freedom
  norb = norb/2 # without spin degree
  fa = open("MX_DISORDER.OUT","w")
  for ib in range(nb): # loop over blocks
    for iorb in range(norb): # loop over orbitals
      wi = (random.random()-0.5)*w  # stregth of the disorder
      hc[ib][ib][2*iorb,2*iorb+1] += wi
      hc[ib][ib][2*iorb+1,2*iorb] += wi
      fa.write(str(ib)+"   "+str(iorb)+"   "+str(wi)+"\n") # save the quantity
  fa.close()


def magnetic_y(hetero,w=0.2):
  """Adds in-plane magnetic disorder, asumes spinpol calculation"""
  hc = hetero.central_intra # this is a nxn tridiagonal list of matrices
  nb = len(hc)  # number of blocks
  norb = len(hc[0][0]) # number of orbitals with spin degree of freedom
  norb = norb/2 # without spin degree
  fa = open("MY_DISORDER.OUT","w")
  for ib in range(nb): # loop over blocks
    for iorb in range(norb): # loop over orbitals
      wi = (random.random()-0.5)*w  # stregth of the disorder
      hc[ib][ib][2*iorb,2*iorb+1] += 1j*wi
      hc[ib][ib][2*iorb+1,2*iorb] += -1j*wi
      fa.write(str(ib)+"   "+str(iorb)+"   "+str(wi)+"\n") # save the quantity
  fa.close()






def random_vacancies(hetero,w=300.0,p_vac=0.0):
  """Adds a random vacancy, asumes spinpol calculation"""
  hc = hetero.central_intra # this is a nxn tridiagonal list of matrices
  nb = len(hc)  # number of blocks
  norb = len(hc[0][0]) # number of orbitals with spin degree of freedom
  fa = open("VACANCY_DISORDER.OUT","w")
  norb = norb/2 # without spin degree
  ind_vac_tot = []  # indexes of the vacancies
  for ib in range(nb): # loop over blocks
    ind_vac = []  # indexes of the vacancies
    for iorb in range(norb): # loop over orbitals
      r = random.random() # random number
      if r<p_vac: # if not chosen yet, add and exit while
        ind_vac.append(iorb)
    ind_vac_tot.append(ind_vac)
    for iorb in ind_vac:
      hc[ib][ib][2*iorb,2*iorb] += w
      hc[ib][ib][2*iorb+1,2*iorb+1] += w
      fa.write(str(ib)+"   "+str(iorb)+"   "+str(1.0)+"\n") # save the quantity
  fa.close()






import numpy as np
from .. import algebra


def get_valley(h,delta=None,**kwargs):
  """Return a callable that calculates the valley expectation value
  using the modified Haldane coupling"""
  if h.dimensionality==0: projector = True # zero dimensional
  ho = h.copy() # copy Hamiltonian
  ho.turn_multicell()
  ho.clean() # set to zero
  ho.add_modified_haldane(1.0/4.5) # add modified Haldane coupling
  hkgen = ho.get_hk_gen() # get generator for the hk Hamiltonian
  def sharpen(m):
    """Sharpen the eigenvalues of a matrix"""
#    return m
    if delta is None: return m # do nothing
    if algebra.issparse(m): return m # temporal workaround
    (es,vs) = algebra.eigh(m) # diagonalize
    es = es/(np.abs(es)+delta) # renormalize the valley eigenvalues
    vs = np.matrix(vs) # convert
    m0 = np.matrix(np.diag(es)) # build new hamiltonian
    return vs@m0@vs.H # return renormalized operator
  def fun(m=None,k=None):
      if h.dimensionality>0 and k is None: raise # requires a kpoint
      hk = hkgen(k) # evaluate Hamiltonian
      hk = sharpen(hk) # sharpen the valley
      if m is None: return hk # just return the valley operator
      else: return hk@m # return the projector
  if h.dimensionality==0: return fun() # return a matrix
  return fun # return function




# different functions specialized for graphene

def add_fwave(h,v,sign=1):
    """Add fwave superconductivity"""
    h0 = h.copy()
    from ..superconductivity import hopping2deltaud
    h0 = h0*0. 
    if sign==1: h0.add_antihaldane(v) 
    elif sign==-1: h0.add_haldane(v) 
    h = hopping2deltaud(h,h0*1j)
    return h





import numpy as np
from .. import klist
from . import pairing

# function to extract different SC order parameters from a Hamiltonian

#
#def fitsc(H,nk):
#    """Fit a Hamiltonian to a collection of uniform SC orders"""
#    hkgen = H.get_hk_gen() # get generator
#    ks = klist.kmesh(nk=nk) # generate klist
#    hks = [] # empty list
#    for k in ks: hks.append(hkgen(k)) # store this matrix
#    # now generate the matrices for all the SC order parameters
#    H0 = H*0. # initialize
#    SCHs = [H0.copy().add_pairing(mode=mode)]
#




import numpy as np
from .. import algebra
from .. import integration
from .rg import green_renormalization
from .kchain import green_kchain



def bloch_selfenergy(h,nk=100,energy = 0.0, delta = 1e-2,
                         mode="full", # algorithm for integration
                         gtype="bulk", # bulk or surface
                         error=1e-3):
  """ Calculates the selfenergy of a cell defect,
      input is a hamiltonian class"""
  if mode=="adaptative": mode = "adaptive"
  def gr(ons,hop):
    """ Calculates G by renormalization"""
    gf,sf = green_renormalization(ons,hop,energy=energy,nite=None,
                            error=error,info=False,delta=delta)
    return gf,sf
  h = h.copy() # make a copy
  h.turn_dense() # dense Hamiltonian
  hk_gen = h.get_hk_gen()  # generator of k dependent hamiltonian
  if h.is_multicell:
      try: h = h.get_no_multicell()
      except:
          mode = "full_adaptive" # multicell hamiltonians only have full mode
          print("Changed to full adaptive mode in selfenergy")
  # sanity check for surface mode
  if gtype=="surface": mode = "adaptive" # only the adaptive mode
  #######################################
  d = h.dimensionality # dimensionality of the system
  g = h.intra *0.0j # initialize green function
  e = np.matrix(np.identity(g.shape[0]))*(energy + delta*1j) # complex energy
  if mode=="full":  # full integration in the BZ
    if d==1: # one dimensional
      ks = [[k,0.,0.] for k in np.linspace(0.,1.,nk,endpoint=False)]
    elif d==2: # two dimensional
      ks = []
      kk = np.linspace(0.,1.,nk,endpoint=False)  # interval 0,1
      for ikx in kk:
        for iky in kk:
          ks.append([ikx,iky,0.])
      ks = np.array(ks)  # all the kpoints
    else: raise # raise error
    for k in ks:  # loop in BZ
      g += algebra.inv(e - hk_gen(k))  # add green function  
    g = g/len(ks)  # normalize
  #####################################################
  #####################################################
  elif mode=="renormalization":
    if d==1: # full renormalization
      g,s = gr(h.intra,h.inter)  # perform renormalization
    elif d==2: # two dimensional, loop over k's
      ks = [[k,0.,0.] for k in np.linspace(0.,1.,nk,endpoint=False)]
#      from ..multicell import rotate90
#      h90 = rotate90(h) # rotated Hamiltonian
      for k in ks:  # loop over k in y direction
 # add contribution to green function
        g += green_kchain(h,k=k,energy=energy,delta=delta,error=error)
#        g += green_kchain(h90,k=k,energy=energy,delta=delta,error=error)
      g = g/len(ks)
    else: raise
  #####################################################
  #####################################################
  elif mode=="adaptive":
    if d==1: # full renormalization
      g,s = gr(h.intra,h.inter)  # perform renormalization
      if gtype=="surface": g = s.copy() # take the surface one
      elif gtype=="bulk": pass # do nothing
      else: raise
    elif d==2: # two dimensional, loop over k's
      ks = [[k,0.,0.] for k in np.linspace(0.,1.,nk,endpoint=False)]
      if gtype=="surface": ig = 1 # take the surface one
      elif gtype=="bulk": ig = 0 # take the bulk one
      else: raise
      def fint(k):
        """ Function to integrate """
        return green_kchain(h,k=[k,0.,0.],energy=energy,
                delta=delta,error=error,only_bulk=False)[ig]
      # eps is error, might work....
      g = integration.integrate_matrix(fint,xlim=[0.,1.],eps=error)
        # chain in the y direction
    else: raise
  elif mode=="full_adaptive":
    fint = lambda k: algebra.inv(e - hk_gen(k))  # green's function
    if d==1: # adaptive 1D
        g = integration.integrate_matrix(fint,xlim=[0.,1.],eps=error)
    elif d==2: # adaptive 2D
        g = integration.integrate_matrix_2D(fint,xlim=[0.,1.],ylim=[0.,1.],
              eps=.1)
    else: raise # not implemented
  # now calculate selfenergy
  selfenergy = e - h.intra - algebra.inv(g)
  return g,selfenergy





import numpy as np
from . import algebra
from .chitk import pmrkky # poor man RKKY



def rkky(h,mode="pm",**kwargs):
    """Compute the RKKY interaction using different methodologies"""
    # this only works with spinless systems
    if mode=="pm": # poor man's methodology
        return pmrkky.explicit_rkky(h,**kwargs) # return explicitly the RKKY
    elif mode=="LR": # linear response theory
        from .chitk import magneticresponse
        return magneticresponse.rkky(h,**kwargs)
    else: raise



def rkky_map(h0,n=2,mode="LR",info=False,fsuper=8,**kwargs):
    """Compute the RKKY map using a brute force algorithm"""
    h = h0.copy() # copy Hamiltonian
    g0 = h.geometry.copy() # copy initial geometry
    if mode=="pm":
        h = h.get_supercell(fsuper*n) # get the supercell Hamiltonian
        r0 = h.geometry.get_closest_position([0.,0.,0.])[0] # central site
        def get_rkky(d,ii,jj): # define routine
            ri = h.geometry.r[ii]
            rj = h.geometry.r[jj]
            d = np.abs(d) # to the fundamental replica
            dr = d[0]*g0.a1 + d[1]*g0.a2 + d[2]*g0.a3 # shift
            e = h.get_rkky(ri=ri,rj=rj+dr,mode="pm",**kwargs) # get the RKKY
            return e
    elif mode=="LR":
        from .chitk.magneticresponse import rkky_generator
        rkkgen = rkky_generator(h,**kwargs) # function to call 
        def get_rkky(d,ii,jj): # define routine
            e = rkkgen(R=d,ii=ii,jj=jj)
#            e = h.get_rkky(R=np.array(d),ii=ii,jj=jj,
#                             mode="LR",**kwargs) # get the RKKY
            if info: print(d,ii,jj,e)
            return e
    else: raise # not implemented

    eout = [] # output RKKY
    rout = [] # output locations
    for d in g0.neighbor_directions(n): # loop over directions
        d = np.array(d)
        jj = 0
        for ii in range(len(g0.r)): # loop over indexes
                if ii==jj and np.max(np.abs(d))==0: continue # same site
                ri = g0.r[ii] # position
                rj = g0.r[jj] # position
                e = get_rkky(d,ii=ii,jj=jj)
                dr = d[0]*g0.a1 + d[1]*g0.a2 + d[2]*g0.a3
                eout.append(e) # store RKKY
                rout.append(dr+ri-rj) # store location   
    rout = np.array(rout) # convert to array
    eout = np.array(eout) # convert to array
    return np.array([rout[:,0],rout[:,1],rout[:,2],eout]).T # return all



import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import minimize
import warnings
warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 
# suppress the numpy warning on arrays


# library to target ingap states


def energy_ingap_state(EO,ne=51,**kwargs):
    """Return the energy of an in-gap state given an embedding object"""
    ev = EO.H.get_gap(mode="valence")
    ec = EO.H.get_gap(mode="conduction")
    es = np.linspace(-ev*0.95,ec*0.95,ne) # number of energies
    delta = (ev+ec)/ne # delta
    (es,ds) = EO.multidos(es=es,delta=delta,**kwargs) # return DOS
    fd = interp1d(np.array(es),-np.array(ds),bounds_error=False,
                         fill_value=(ds[0],ds[-1]),
                        kind="quadratic")
    res = minimize(fd,np.array([0.0]),bounds=np.array([[min(es),max(es)]]),
                          method="Powell")
    return res.x[0]




import numpy as np



def bloch_phase(self,d,k):
    """
    Return the Bloch phase for this d vector
    """
    if self.dimensionality == 0: return 1.0
    elif self.dimensionality == 1:
      try: kp = k[0] # extract the first component
      except: kp = k # ups, assume that it is a float
      dt = np.array(d)[0]
      kt = np.array([kp])[0]
      return np.exp(1j*dt*kt*np.pi*2.)
    elif self.dimensionality == 2:
      dt = np.array(d)[0:2]
      kt = np.array(k)[0:2]
      return np.exp(1j*dt.dot(kt)*np.pi*2.)
    elif self.dimensionality == 3:
      dt = np.array(d)[0:3]
      kt = np.array(k)[0:3]
      return np.exp(1j*dt.dot(kt)*np.pi*2.)
    else: raise






import numpy as np
from .. import algebra
from numba import jit

use_fortran = False
use_numba = False

def green_renormalization(intra,inter,**kwargs):
    if use_numba:
      return green_renormalization_jit(intra,inter,**kwargs)
    else: 
      return green_renormalization_python(intra,inter,**kwargs)




def green_renormalization_python(intra,inter,energy=0.0,nite=None,
                            info=False,delta=0.001,
                            use_fortran = use_fortran,**kwargs):
    """ Calculates bulk and surface Green function by a renormalization
    algorithm, as described in I. Phys. F: Met. Phys. 15 (1985) 851-858 """
    intra = algebra.todense(intra)
    inter = algebra.todense(inter)
    error = np.abs(delta)*1e-6 # overwrite error
    e = np.matrix(np.identity(intra.shape[0])) * (energy + 1j*delta)
    ite = 0
    alpha = inter.copy()
    beta = algebra.dagger(inter).copy()
    epsilon = intra.copy()
    epsilon_s = intra.copy()
    while True: # implementation of Eq 11
      einv = algebra.inv(e - epsilon) # inverse
      epsilon_s = epsilon_s + alpha @ einv @ beta
      epsilon = epsilon + alpha @ einv @ beta + beta @ einv @ alpha
      alpha = alpha @ einv @ alpha  # new alpha
      beta = beta @ einv @ beta  # new beta
      ite += 1
      # stop conditions
      if not nite is None:
        if ite > nite:  break
      else:
        if np.max(np.abs(alpha))<error and np.max(np.abs(beta))<error: break
    if info:
      print("Converged in ",ite,"iterations")
    g_surf = algebra.inv(e - epsilon_s) # surface green function
    g_bulk = algebra.inv(e - epsilon)  # bulk green function
    return g_bulk,g_surf

def green_renormalization_jit(intra,inter,energy=0.0,delta=1e-4,**kwargs):
    intra = algebra.todense(intra)*(1.0+0j)
    inter = algebra.todense(inter)*(1.0+0j)
    g0 = intra*0.0j
    g1 = intra*0.0j
    nite = int(10/delta)
    error = delta*1e-3
    energyz = energy + 1j*delta
    e = np.array(np.identity(intra.shape[0]),dtype=np.complex) * energyz
    return green_renormalization_jit_core(g0,g1,intra,inter,e,nite,
                                                error)


#@jit()
#@jit(nopython=True)
### this seems to not work with numba ###
def green_renormalization_jit_core(g0,g1,intra,inter,e,nite,error):
    ite = 0
    alpha = inter*1.0
    beta = np.conjugate(inter).T*1.0
    epsilon = intra*1.0
    epsilon_s = intra*1.0
    while True: # implementation of Eq 11
      einv = np.linalg.inv(e - epsilon) # inverse
      epsilon_s = epsilon_s + alpha @ einv @ beta
      epsilon = epsilon + alpha * einv @ beta + beta @ einv @ alpha
      alpha = alpha @ einv @ alpha  # new alpha
      beta = beta @ einv @ beta  # new beta
      ite += 1
      # stop conditions
      if np.max(np.abs(alpha))<error and np.max(np.abs(beta))<error: break
    g_surf = np.linalg.inv(e - epsilon_s) # surface green function
    g_bulk = np.linalg.inv(e - epsilon)  # bulk green function 
    g0,g1 = g_bulk*1,g_surf*1
    return g0,g1
                     



import numpy as np
from scipy.integrate import quad

thermalmode = "adaptive" # thermal mode

def finite_T_didv(self,temp,energy=0.0,**kwargs):
    """Finite temperature dIdV"""
    from .didv import zero_T_didv
    if thermalmode=="adaptive":
        from .fermidirac import fermidirac as FD
        dt = 20 # max T range
        de = temp # energy difference to compute the derivative
        ### Use simpson integration
        def f(e):
            out = zero_T_didv(self,energy=energy+e,**kwargs)
            out *= FD(e-de,temp=temp) - FD(e+de,temp=temp)
            return out/de
        from ..integration import peak_integrate
        return quad(f,-dt*temp,dt*temp,epsrel=1e-4,limit=60)[0]/2.
    elif thermalmode=="pm": # poor's man mode
        from .fermidirac import dFDdT
        nT = 45
        Ts = np.linspace(-4*temp,4*temp,nT)
        def f(e):
            return zero_T_didv(self,energy=energy+e,**kwargs)
        out = [f(e)*dFDdT(e,temp=temp)*temp for e in Ts]
        norm = np.sum(dFDdT(Ts,temp=temp)*temp)
        out = np.trapz(out)/norm
        return out
    else: raise





from __future__ import print_function
import numpy as np
from copy import deepcopy
from scipy.sparse import bmat
from scipy.sparse import csc_matrix as csc
from . import sculpt
from . import klist
from .supercell import non_orthogonal_supercell
from . import supercell as supercelltk
from . import checkclass
import scipy.linalg as lg
from numba import jit
from .htk.g2h import get_hamiltonian
from .helptk import get_docstring

use_fortran = False

class Geometry:
    """ Class for a geometry in a system """
    def __init__(self):
        self.data = dict() # empty dictionary with different data
        self.has_sublattice = False # has sublattice index
        self.sublattice_number = 2 # two sublattices
        self.has_fractional = False
        self.dimensionality = 1 # dimension of the hamiltonian
        self.x = [] # positions in x
        self.y = [] # positions in y
        self.z = [] # positions in z
        self.r = [] # full positions 
        self.celldis = 1.0 # distance to the nearest cell (for 1d)
        self.a1 = np.array([100.0,0.0,0.])  # first vector to the nearest cell
        self.a2 = np.array([0.0,100.0,0.])  # first vector to the nearest cell
        self.a3 = np.array([0.0,0.0,100.])  # first vector to the nearest cell
        self.b1 = np.array([1.0,0.0,0.])  # first vector to the nearest cell
        self.b2 = np.array([0.0,1.0,0.])  # first vector to the nearest cell
        self.b3 = np.array([0.0,0.0,1.])  # first vector to the nearest cell
        self.shift_kspace = False # shift the klist when plotting
        self.name = "None"
        self.primal_geometry = None # store the primal geometry
        self.lattice_name = "" # lattice name
        self.atoms_names = [] # no name for the atoms
        self.atoms_have_names = False # atoms do not have names
        self.ncells = 2 # number of neighboring cells returned
    def neighbor_distances(self,**kwargs):
        return neighbor_distances(self,**kwargs)
    def get_neighbor_distances(self,**kwargs):
        return neighbor_distances(self,**kwargs)
    def normalize_nn_distance(self):
        """Set the NN istance equal to 1"""
        if self.dimensionality>0: raise
        d = self.neighbor_distances(n=1)[0]
        self.r = self.r/d
        self.r2xyz()
    def get_index(self,r,**kwargs):
        return get_index(self,r,**kwargs)
    def __add__(self,g1):
        from .geometrytk.galgebra import sum_geometries
        return sum_geometries(self,g1)
    def __sub__(self,a):
        return self + (-1)*a
    def __radd__(self,g1):
        return sum_geometries(self,g1)
    def get_kmesh(self,**kwargs):
        """Return the k-mesh"""
        return klist.kmesh(self.dimensionality,**kwargs)
    def get_default_kpath(self,**kwargs):
        from . import klist
        return klist.default(self,**kwargs)
    def set_finite(self,periodic=False):
      """ Transfrom the geometry into a finite system"""
      if periodic:
        f = self.periodic_vector() # get the function
        self.get_distance = f # store that function
      self.dimensionality = 0 # set as finite
    def get_orthogonal(self):
        return supercelltk.target_angle_volume(self,angle=0.5)
    def closest_index(self,r):
        return sculpt.get_closest(self,n=1,r0=r)[0]
    def get_closest_position(self,r,n=1):
        r = np.array(r)
        if n==0:
          ii = self.closest_index(r)
          return self.r[ii] # return this position
        else:
          iis = sculpt.get_closest(self,n=n,r0=r)
          return [self.r[ii] for ii in iis] # return positions
    def get_supercell(self,nsuper,**kwargs):
        return get_supercell(self,nsuper,**kwargs)
    supercell = get_supercell # backwards compatibility
    def xyz2r(self):
      """Updates r atributte according to xyz"""
      self.r = np.array([self.x,self.y,self.z]).transpose()
    def r2xyz(self):
      """Updates x,y,z atributtes according to r"""
      r = np.array(self.r).transpose()
      self.x = r[0]
      self.y = r[1]
      self.z = r[2]
    @get_docstring(get_hamiltonian) # inherint docstring
    def get_hamiltonian(self,**kwargs):
        return get_hamiltonian(self,**kwargs)
    def write(self,**kwargs):
        """ Writes the geometry in file"""
        write_positions(self,**kwargs)
        write_xyz(self)
        write_lattice(self)
        write_sublattice(self)
    def get_kpath(self,*args,**kwargs):
        return klist.get_kpath(self,*args,**kwargs)
    def write_positions(self,**kwargs):
        """Write the positions in a file"""
        write_positions(self,**kwargs)
    def copy(self):
        """Copy the geometry"""
        return deepcopy(self)
    def set_origin(self,r=None):
        if r is None: r = self.r[self.get_central()[0]]
        self.x = self.x - r[0]
        self.y = self.y - r[1]
        self.z = self.z - r[2]
        self.xyz2r() # update r
    def center(self):
        """ Centers the geometry in (0,0,0)"""
        self.x = self.x - sum(self.x)/len(self.x)
        self.y = self.y - sum(self.y)/len(self.y)
        self.z = self.z - sum(self.z)/len(self.z)
        self.xyz2r() # update r
    def get_lattice_name(self):
        if self.dimensionality==2:
            if np.abs(self.a1.dot(self.a2))<0.0001:        
              self.lattice_name = "square"
            else:
              self.lattice_name = "triangular"
    def get_k2K(self):
        from .kpointstk.mapping import get_k2K
        return get_k2K(self)
    def reciprocal2natural(self,v):
        """
        Return a natural vector in real reciprocal coordinates
        """
        return self.get_k2K_generator()(v)
    def get_fractional(self,center=False):
        """Fractional coordinates"""
        self.update_reciprocal() # update reciprocal lattice vectors
        get_fractional(self,center=center) # get fractional coordinates
    def rotate(self,angle):
      """Rotate the geometry"""
      return sculpt.rotate(self,angle*np.pi/180)
    def clean(self,iterative=False):
      return sculpt.remove_unibonded(self,iterative=iterative)
    def get_diameter(self):
      """Return the maximum distance between two atoms"""
      return get_diameter(self)  
    def periodic_vector(self):
      return periodic_vector(self)
    def get_sublattice(self):
      """Initialize the sublattice"""
      if self.has_sublattice: self.sublattice = get_sublattice(self.r)
      else: 
          self.sublattice = get_sublattice(self.r)
          self.has_sublattice = True
    def shift(self,r0):
      """Shift all the positions by r0"""
      self.x[:] -= r0[0]
      self.y[:] -= r0[1]
      self.z[:] -= r0[2]
      self.xyz2r() # update
      if self.dimensionality>0:
        self.get_fractional(center=True)
        self.fractional2real()
    def write_function(self,fun,**kwargs):
        from .geometrytk.write import write_function
        return write_function(self,fun,**kwargs)
    def neighbor_directions(self,n=None):
      """Return directions linking to neighbors"""
      if n is None: n = self.ncells
      return neighbor_directions(self,n)
    def get_ncells(self):
        if self.dimensionality==0: return 0
        else:
            n = int(10/np.sqrt(self.a1.dot(self.a1)))
            return max([1,n])
    def write_profile(self,d,**kwargs):
        """Write a profile in a file"""
        write_profile(self,d,**kwargs)
    def replicas(self,**kwargs):
        from .geometrytk.replicas import replicas
        return replicas(self,**kwargs)
    def multireplicas(self,n):
        from .geometrytk.replicas import multireplicas
        return multireplicas(self,n)
    def bloch_phase(self,d,k):
        """Return the Bloch's phase for a specific k-vector"""
        from .geometrytk.bloch import bloch_phase
        return bloch_phase(self,d,k)
    def remove(self,i=0):
        """
        Remove one site
        """
        if callable(i): return sculpt.intersec(self,lambda r: not i(r))
        if type(i)==list: pass
        else: i = [i]
        return sculpt.remove(self,i)
    def center_in_atom(self,n0=None):
        """
        Center the geometry in an atom
        """
        if n0 is None: n0 = sculpt.get_central(self)[0] # get the index
        sculpt.shift(self,r=self.r[n0]) # shift the geometry
    def get_central(self,n=1):
        """
        Return a list of central atoms
        """
        return sculpt.get_central(self,n=n) # get the index
    def update_reciprocal(self):
        """
        Update reciprocal lattice vectors
        """
        self.b1,self.b2,self.b3 = get_reciprocal(self.a1,self.a2,self.a3)
    def get_k2K_generator(self,**kwargs):
        return get_k2K_generator(self,**kwargs)
    def k2K(self,k): return get_k2K_generator(self,toreal=False)(k)
    def K2k(self,k): return get_k2K_generator(self,toreal=True)(k)
    def fractional2real(self):
      """
      Convert fractional coordinates to real coordinates
      """
      fractional2real(self)
    def real2fractional(self):
      self.get_fractional() # same function
    def add_strain(self,*args,**kwargs):
        from .geometrytk import strain
        return strain.add_strain(self,*args,**kwargs)
    def get_connections(self):
      """
      Return the connections of each site
      """
      from . import neighbor
      self.connections = neighbor.connections(self.r,self.r)
      return self.connections # return list



def honeycomb_armchair_ribbon(ntetramers=10):
  """ Creates the positions of an armchair ribbon
  of width ntetramers, return a geometry class """
  from numpy import array, sqrt
  n = ntetramers
  x=array([0.0 for i in range(4*n)])
  y=array([0.0 for i in range(4*n)])
  s3=sqrt(3.0)/2.0 
  for ii in range(n):
    fi=float(ii)*s3*2.0
    i=4*ii
    x[i]=0.0
    x[i+1]=1.0
    x[i+2]=1.5
    x[i+3]=2.5 
    y[i]=fi
    y[i+1]=fi
    y[i+2]=fi+s3
    y[i+3]=fi+s3
  x=x-sum(x)/float(4*n)
  y=y-sum(y)/float(4*n)
  g = Geometry() # create geometry class
  g.x = x  # add to the x atribute
  g.y = y  # add to the y atribute
  g.z = y*0.0  # add to the y atribute
  g.celldis = 3.0 # add distance to the nearest cell
  g.a1 = np.array([3.0,0.,0.]) # add distance to the nearest cell
  g.shift_kspace = True # shift kpoint when plotting
  g.has_sublattice = True # has sublattice index
  g.sublattice = [(-1.)**i for i in range(len(x))] # subattice number
  g.name = "honeycomb_armchair_ribbon"  # name of the geometry
  g.xyz2r() # create r coordinates
  g.dimensionality = 1
  return g


honeycomb_ribbon = honeycomb_armchair_ribbon # alias



def square_ribbon(natoms):
  """ Creates the hamiltonian of a square ribbon lattice"""
  from numpy import array
  x=array([0.0 for i in range(natoms)]) # create x coordinates
  y=array([float(i) for i in range(natoms)])  # create y coordinates
  y=y-np.sum(y)/float(natoms) # shift to the center
  g = Geometry() # create geometry class
  g.x = x  # add to the x atribute
  g.y = y  # add to the y atribute
  g.z = y*0.0  # add to the y atribute
  g.celldis = 1.0 # add distance to the nearest cell
  g.a1 = np.array([1.0,0.,0.]) # add distance to the nearest cell
  g.xyz2r() # create r coordinates
  g.has_sublattice = False # does not have sublattice
  return g


def ladder(): return square_ribbon(2)

def bisquare_ribbon(ncells):
  g = square_lattice_bipartite()
  g = g.supercell((1,ncells))
  g.dimensionality = 1
  return g



def chain(n=1):
  """ Create a chain """
  g = square_ribbon(1) 
  g = g.get_supercell(n)
  g.has_sublattice = False
  g.get_fractional()
#  g.sublattice = [(-1)**i for i in range(len(g.x))]
  return g



def bichain(n=1):
    """ Create a chain """
    g = square_ribbon(1) 
    g = g.get_supercell(2)
    g.has_sublattice = True
    g.sublattice = [(-1)**i for i in range(len(g.x))]
    g = g.get_supercell(n)
    return g


def dimer():
    """Geomtry of a dimer"""
    g = bichain()
    g.dimensionality = 0
    return g




def square_tetramer_ribbon(ntetramers):
    """ Creates the hamiltonian of a square tetramer ribbon lattice"""
    from numpy import array
    natoms = ntetramers*4
    x=array([0.0 for i in range(natoms)]) # create x coordinates
    y=array([0.0 for i in range(natoms)])  # create y coordinates
    for i in range(ntetramers):
      x[4*i] = 0.0
      x[4*i+1] = 1.0
      x[4*i+2] = 1.0
      x[4*i+3] = 0.0
      y[4*i] = 2.*i
      y[4*i+1] = 2.*i
      y[4*i+2] = 2.*i +1.0
      y[4*i+3] = 2.*i +1.0
    y=y-sum(y)/float(natoms) # shift to the center
    x=x-sum(x)/float(natoms) # shift to the center
    g = Geometry() # create geometry class
    g.x = x  # add to the x atribute
    g.y = y  # add to the y atribute
    g.z = y*0.  # add to the z atribute
    g.celldis = 2.0 # add distance to the nearest cell
    g.a1 = np.array([2.0,0.,0.]) # add distance to the nearest cell
    g.shift_kspace = True # add distance to the nearest cell
    g.xyz2r() # create r coordinates
    g.has_sublattice = True # has sublattice index
    g.sublattice = [(-1.)**i for i in range(len(x))] # subattice number
    g.dimensionality = 1
    return g







def square_zigzag_ribbon(npairs):
  """ Creates the hamiltonian of a square zigzag (11) lattice"""
  from numpy import array,sqrt
  s2 = sqrt(2.) # square root of 2
  natoms = 2*npairs
  x=array([s2/4.*(-1)**i for i in range(natoms)]) # create x coordinates
  y=array([0.0 for i in range(natoms)])  # create y coordinates of pairs
  yp=array([s2*float(i) for i in range(npairs)])  # create y coordinates of pairs
  for i in range(npairs): # y position in each pair
    y[2*i] = yp[i]
    y[2*i+1] = yp[i] + s2/2.
  y=y-sum(y)/float(natoms) # shift to the center
  g = Geometry() # create geometry class
  g.x = x  # add to the x atribute
  g.y = y  # add to the y atribute
  g.celldis = s2 # add distance to the nearest cell
  g.xyz2r() # create r coordinates
  g.dimensionality = 1
  return g




def honeycomb_zigzag_ribbon(ntetramers=10):
  from numpy import array, sqrt
  n = ntetramers
  x=array([0.0 for i in range(4*n)])
  y=array([0.0 for i in range(4*n)])
  s3=sqrt(3.0)/2.0
  for ii in range(n):
    fi=-float(ii)*3.0
    i=4*ii
    x[i]=0.0
    x[i+1]=s3
    x[i+2]=s3
    x[i+3]=0.0
    y[i]=fi
    y[i+1]=fi-0.5
    y[i+2]=fi-1.5
    y[i+3]=fi-2.0
  x=x-sum(x)/float(4*n)
  y=y-sum(y)/float(4*n)
  g = Geometry() # create geometry class
  g.x = x  # add to the x atribute
  g.y = y  # add to the y atribute
  g.z = y*0.0  # add to the z atribute
  g.celldis = sqrt(3.0) # add distance to the neares cell
  g.a1 = np.array([sqrt(3.0),0.,0.]) # add distance to the nearest cell
  g.has_sublattice = True # has sublattice index
  g.sublattice = [(-1.)**i for i in range(len(x))] # subattice number
  g.name = "honeycomb_zigzag_ribbon"
  g.xyz2r() # create r coordinates
  g.dimensionality = 1
  return g



def honeycomb_lattice_zigzag():
  """ Return a honeyomb lattice with 4 atoms per unit cell"""
  from numpy import array, sqrt
  x=array([0.0 for i in range(4)])
  y=array([0.0 for i in range(4)])
  s3=sqrt(3.0)/2.0
  for ii in range(1):
    fi=-float(ii)*3.0
    i=4*ii
    x[i]=0.0
    x[i+1]=s3
    x[i+2]=s3
    x[i+3]=0.0
    y[i]=fi
    y[i+1]=fi-0.5
    y[i+2]=fi-1.5
    y[i+3]=fi-2.0
  g = Geometry() # create geometry class
  g.x = x  # add to the x atribute
  g.y = -y  # add to the y atribute
  g.z = y*0.0  # add to the z atribute
  g.a1 = np.array([sqrt(3.0),0.,0.]) # a1 distance
  g.a2 = np.array([0.,3.,0.]) # a1 distance
  g.has_sublattice = True # has sublattice index
  g.sublattice = [(-1.)**i for i in range(len(x))] # subattice number
  g.name = "honeycomb_zigzag_ribbon"
  g.xyz2r() # create r coordinates
  g.dimensionality = 2
  g.center()
  return g

def honeycomb_lattice_armchair():
    g = honeycomb_lattice_zigzag()
    g.a1,g.a2 = g.a2,-g.a1 # switch axis
    from . import sculpt
    g = sculpt.rotate_a2b(g,g.a1,np.array([1.0,0.0,0.0]))
    return g


def supercell1d(g,nsuper):
  """
  Creates a supercell of the system
  """
  # get the old geometry 
  y = g.y
  x = g.x
  z = g.z
  celldis = g.a1[0]
  if np.abs(g.a1.dot(g.a1) - g.a1[0]**2)>0.001:
    print("Something weird in supercell 1d")
    return supercell1d(sculpt.rotate_a2b(g,g.a1,np.array([1.,0.,0.])),nsuper) 
  # position of the supercell
  yout = []
  xout = []
  for i in range(nsuper):
    yout += y.tolist()
    xout += (x+i*celldis).tolist()
  # now modify the geometry
  go = deepcopy(g)
  go.x = np.array(xout)
  go.y = np.array(yout)
  # and shift to zero
  go.z = np.array(z.tolist()*nsuper)
  go.center() # center the unit cell
  go.celldis = celldis*nsuper
  go.a1 = g.a1*nsuper # supercell
  go.xyz2r() # update r
  if g.has_sublattice: # if has sublattice, keep the indexes
    go.sublattice = np.concatenate([g.sublattice for i in range(nsuper)])  # store the keeped atoms
#    print(nsuper)
  if g.atoms_have_names: # supercell sublattice
    go.atoms_names = g.atoms_names*nsuper
  go.get_fractional()
  return go






################################################
########### begin 2d geometries ################
################################################

def honeycomb_lattice(n=1):
  """
  Create a honeycomb lattice
  """
  g = Geometry() # create geometry
  g.x = np.array([-0.5,0.5])
  g.y = np.array([0.0,0.0])
  g.z = np.array([0.0,0.0])
  g.a1 = np.array([3./2.,np.sqrt(3.)/2,0.]) # first lattice vector
  g.a2 = np.array([-3./2.,np.sqrt(3.)/2,0.]) # second lattice vector
  g.a3 = np.array([0.,0.,10.]) # second lattice vector
  g.dimensionality = 2 # two dimensional system
  g.xyz2r() # create r coordinates
  g.has_sublattice = True # has sublattice index
  g.sublattice = [(-1.)**i for i in range(len(g.x))] # subattice number
  g.update_reciprocal() # update reciprocal lattice vectors
  g.get_fractional()
  if n>1: return supercelltk.target_angle(g,angle=1./3.,volume=int(n),
          same_length=True) 
  return g


def buckled_honeycomb_lattice(n=1):
  """
  Return a buckled honeycomb lattice
  """
  from . import films
  g = diamond_lattice_minimal()
  g = films.geometry_film(g,nz=n)
  return g 



def triangular_lattice(n=1):
  """
  Creates a triangular lattice
  """
  g = Geometry() # create geometry
  g.x = np.array([0.0])
  g.y = np.array([0.0])
  g.z = np.array([0.0])
  g.a1 = np.array([np.sqrt(3.)/2.,1./2,0.]) # first lattice vector
  g.a2 = np.array([-np.sqrt(3.)/2.,1./2,0.]) # second lattice vector
  g.dimensionality = 2 # two dimensional system
  g.xyz2r() # create r coordinates
  g.has_sublattice = False # has sublattice index
  g.get_fractional() # update reciprocal lattice vectors
  if n>1: return supercelltk.target_angle_volume(g,angle=1./3.,volume=int(n),
          same_length=True) 
  g = sculpt.rotate_a2b(g,g.a1,np.array([1.,0.,0.]))
  return g




def triangular_lattice_tripartite():
  """
  Creates a triangular lattice with three sites per unit cell
  """
  g = triangular_lattice()
  return supercelltk.target_angle_volume(g,angle=1./3.,volume=3,
          same_length=True)



def triangular_lattice_pentapartite():
  """
  Creates a triangular lattice with five sites per unit cell
  """
  g = triangular_lattice()
  return supercelltk.target_angle(g,angle=1./3.,volume=5,same_length=True)





def triangular_ribbon(n):
  g = triangular_lattice() # create geometry
  go = g.copy() # copy geometry
  r0 = [] # empty list
  for ir in g.r:
    r0.append(ir) # supercell
    r0.append(ir+g.a1) # supercell
  rs = []
  dr = g.a1+g.a2 # displacement vector
  for i in range(n): # loop over replicas
    for ir in r0: # loop over unit cell
      rs.append(dr*i + ir) # append atom
  go.r = np.array(rs) # save coordinates
  go.r2xyz() # update
  go.a1 = g.a1 - g.a2 #
  go.center()
  go.dimensionality = 1
  # now rotate the geometry
  from . import sculpt
  go = sculpt.rotate_a2b(go,go.a1,np.array([1.0,0.0,0.0]))
  # setup the cell dis parameter (deprecated)
  go.celldis = go.a1[0]
  return go



def square_lattice_bipartite():
  """
  Creates a square lattice
  """
  g = Geometry() # create geometry
  g.x = np.array([-0.5,0.5,0.5,-0.5])
  g.y = np.array([-0.5,-0.5,0.5,0.5])
  g.z = g.x*0.
  g.a1 = np.array([2.,0.,0.]) # first lattice vector
  g.a2 = np.array([0.,2.,0.]) # second lattice vector
  g.dimensionality = 2 # two dimensional system
  g.xyz2r() # create r coordinates
  g.has_sublattice = True # has sublattice index
  g.sublattice = [(-1.)**i for i in range(len(g.x))] # subattice number
  g.update_reciprocal() # update reciprocal lattice vectors
  return g



def square_lattice():
  """
  Creates a square lattice
  """
  g = Geometry() # create geometry
  g.x = np.array([0.])
  g.y = np.array([0.])
  g.z = g.x*0.
  g.a1 = np.array([1.,0.,0.]) # first lattice vector
  g.a2 = np.array([0.,1.,0.]) # second lattice vector
  g.dimensionality = 2 # two dimensional system
  g.xyz2r() # create r coordinates
  g.has_sublattice = False # has sublattice index
  g.sublattice = [0. for i in range(len(g.r))] # subattice number
  return g


single_square_lattice = square_lattice # alias



def cubic_lattice():
  """
  Creates a cubic lattice
  """
  g = Geometry() # create geometry
  g.r = [np.array([0.,0.,0.])]
  g.x = [0.0]
  g.y = [0.0]
  g.z = [0.0]
  g.a1 = np.array([1.,0.,0.]) # first lattice vector
  g.a2 = np.array([0.,1.,0.]) # second lattice vector
  g.a3 = np.array([0.,0.,1.]) # second lattice vector
  g.dimensionality = 3 # three dimensional system
  g.has_sublattice = False # has sublattice index
  g.sublattice = [0. for i in range(len(g.r))] # subattice number
  return g


cubic_lattice_minimal = cubic_lattice


def cubic_lattice_bipartite():
  """
  Creates a cubic lattice
  """
  g = Geometry() # create geometry
  a1 = np.array([1.,0.,0.]) # first lattice vector
  a2 = np.array([0.,1.,0.]) # second lattice vector
  a3 = np.array([0.,0.,1.]) # second lattice vector
  rs = []
  ss = []
  for i in range(2):
    for j in range(2):
      for k in range(2):
        ss.append((-1)**(i+j+k)) # sublattice
        rs.append(i*a1 + j*a2 + k*a3) # position
  g.a1 = a1*2      
  g.a2 = a2*2      
  g.a3 = a3*2      
  g.sublattice = np.array(ss)
  g.r = np.array(rs)
  g.r2xyz()
  g.dimensionality = 3 # three dimensional system
  g.has_sublattice = True # has sublattice index
#  g.sublattice = [(-1.)**i for i in range(len(g.x))] # subattice number
  return g


def cubic_lieb_lattice():
  """
  Return a 3d Lieb lattice
  """
  g = cubic_lattice_bipartite()
  g = g.remove(0) # remove this atom
  return g


def lieb_lattice():
  """
  Create a 2d Lieb lattice
  """
  g = Geometry() # create geometry
  g.x = np.array([-0.5,0.5,0.5])
  g.y = np.array([-0.5,-0.5,0.5])
  g.z = g.x*0.
  g.a1 = np.array([2.,0.,0.]) # first lattice vector
  g.a2 = np.array([0.,2.,0.]) # second lattice vector
  g.dimensionality = 2 # two dimensional system
  g.xyz2r() # create r coordinates
  g.has_sublattice = True # has sublattice index
  g.sublattice = [(-1.)**i for i in range(len(g.x))] # subattice number
  return g




def kagome_lattice(n=1):
  """
  Creates a Kagome lattice
  """
  g = Geometry() # create geometry
  dx = 1./2.
  dy = np.sqrt(3)/2.
  g.x = np.array([-dx,dx,0.])
  g.y = np.array([-dy,-dy,0.0])
  g.z = np.array([0.0,0.0,0.])
  g.a1 = np.array([2.,0.,0.]) # first lattice vector
  g.a2 = np.array([1.,np.sqrt(3),0.]) # second lattice vector
  g.dimensionality = 2 # two dimensional system
  g.xyz2r() # create r coordinates
  g.center()
  g.has_sublattice = True # does not have sublattice index
  g.sublattice_number = 3 # three sublattices
  g.sublattice = [0,1,2] # the three sublattices
  if n>1: return supercelltk.target_angle(g,angle=1./3.,volume=int(n),
          same_length=True) 
  g.update_reciprocal()
  return g


def rectangular_kagome_lattice():
  """
  Creates a square kagome lattice
  """
  g = kagome_lattice()
  g = g.supercell(2) # create a supercell
  go = g.copy()
  go.a2 = np.array([0.0,np.sqrt(3)*2.0,0.0]) 
  return go







def honeycomb_lattice_square_cell():
  """
  Creates a honeycomb lattice
  """
  from .supercell import target_angle_volume
  g = honeycomb_lattice() # create geometry
  return target_angle_volume(g,volume=2,angle=.5)



def honeycomb_lattice_C6():
  """
  Geometry for a honeycomb lattice, taking a unit cell
  with C6 rotational symmetry
  """
  g = honeycomb_lattice() # create geometry
  return supercelltk.target_angle_volume(g,angle=1./3.,volume=3,
          same_length=True)







def kagome_ribbon(n=5):
  """Create a Kagome ribbon"""
  g = rectangular_kagome_lattice() # 2d geometry
  from . import ribbonizate
  g = ribbonizate.bulk2ribbon(g,n=n) # create ribbon from 2d
  return g




from .supercell import supercell2d
from .supercell import supercell3d




def read(input_file="POSITIONS.OUT"):
  """ Reads a geometry """
  m = np.genfromtxt(input_file).transpose()
  g = Geometry() # cretae geometry
  g.dimensionality = 0
  g.x = m[0]
  g.y = m[1]
  g.z = m[2]
  g.xyz2r() # create r coordinates
  # check whether if it has sublattice
  try:
    g.sublattice = np.genfromtxt("SUBLATTICE.OUT")
    g.has_sublattice = True
    if len(g.sublattice) != len(g.r): raise
    print("Read sublattice from SUBLATTICE.OUT")
  except: g.has_sublattice = False
  try:
    dim = int(open("DIMENSIONALITY.OUT").read())
  except: dim = 0
  g.dimensionality = dim # store
  if dim>0: # if it has lattice
    lat = np.genfromtxt("LATTICE.OUT")   # read lattice
    if dim==2: # two dimensional
      g.a1 = lat[0]
      g.a2 = lat[1]
    elif dim==3: # two dimensional
      g.a1 = lat[0]
      g.a2 = lat[1]
      g.a3 = lat[2]
    elif dim==1: # two dimensional
      g.celldis = np.sqrt(lat.dot(lat)) # distance between cells
      g.a1 = lat
  return g




from .ribbon import bulk2ribbon
#def bulk2ribbon(h,n=10):
#  """Converts a hexagonal bulk hamiltonian into a ribbon hamiltonian"""
#  from .hexagonal import bulk2ribbon as br
#  return br(h,n=n)


def get_reciprocal2d(a1,a2):
  """Get reciprocal vectors"""
  b1 = np.array([a2[1],-a2[0],0.])
  b2 = np.array([-a1[1],a1[0],0.])
  b1 = b1 / np.sqrt(b1.dot(b1))
  b2 = b2 / np.sqrt(b2.dot(b2))
  return (b1,b2)


def apilate(g,drs=[np.array([0.,0.,0.])]):
  """ generate a geometry by appilating another geometry, displacements
  are given on input """
  nl = len(drs) # number of layers
  ro = np.concatenate([g.r + dr for dr in drs])
  go = g.copy()
  go.r = ro
  go.r2xyz()
  return go



from .geometrytk.write import write_xyz
from .geometrytk.write import write_lattice
from .geometrytk.write import write_sublattice
from .geometrytk.write import write_positions
write_geometry = write_positions


def remove_duplicated(g):
  """ Remove duplicated atoms"""
  if not g.atoms_have_names: raise
  go = g.copy() # copy geometry
  rs = remove_duplicated_positions(g.r)
  go.r = np.array(rs)
  go.r2xyz() # update the other coordinates
#  go.atoms_names = names
  return go


def remove_duplicated_positions(r):
  rs = []
  for ir in r: # loop over atoms
     store = True # store this atom
     for jr in rs: # loop over stored
       dr = ir-jr
       dr = dr.dot(dr) # distance
       if dr<0.01: store = False
     if store: # store this atom
       rs.append(ir.copy())
  return np.array(rs) # return unrepeated atoms




def get_reciprocal(a1,a2,a3):
  """Return the reciprocal lattice vectors
  By definition, ai*bj = delta_ij"""
  (ux,uy,uz) = (a1,a2,a3)
#  ux = ux/np.sqrt(ux.dot(ux))
#  uy = uy/np.sqrt(uy.dot(uy))
#  uz = uz/np.sqrt(uz.dot(uz))
  a2kn = np.matrix([ux,uy,uz]) # matrix for the change of basis
  r2a = np.matrix([ux,uy,uz]).T.I # from real space to lattice vectors
  b1,b2,b3 = r2a[0,:],r2a[1,:],r2a[2,:]
  b1 = np.array(b1).reshape(3)
  b2 = np.array(b2).reshape(3)
  b3 = np.array(b3).reshape(3)
  return b1,b2,b3


from .geometrytk.fractional import get_fractional_function
from .geometrytk.fractional import get_fractional
from .geometrytk.fractional import fractional2real





def get_diameter(g):
  """Get the maximum distance between atoms"""
  from scipy.spatial import distance
  r = np.array(g.r) # positions
  dis = distance.cdist(r,r) # distances
  return np.max(dis)
  


def periodic_vector(g):
  """Returns a function that calculates the distance between
  two sites, using periodic boundary conditions"""
  if g.dimensionality != 2:
    print("WARNING, not 2d")
  a1 = g.a1 # first unit vector
  a2 = g.a2 # second unit vector
  shifts = [] # empty list
  for i in [-1,0,1]: # loop over closest cells
    for j in [-1,0,1]: # loop over closest cells
      shifts.append(i*a1+j*a2) # list of possible vectors
  def dis(r1,r2):
    dr = r1-r2
    rr = [dr + v for v in shifts] # loop over shifts
    rr2 = [r.dot(r) for r in rr] # distance
    mr = np.min(rr2)
    return rr[rr2.index(mr)] # return minimum
  return dis # return function
 



def periodic_zrotation(g,v=np.array([0.,0.])):
  """Returns a function that calculates the rotation between
  two sites, applying twisted boundary conditions"""
  if g.dimensionality != 2:
    print("WARNING, not 2d")
  a1 = g.a1 # first unit vector
  a2 = g.a2 # second unit vector
  shifts = [] # empty list
  index = [] # empty list
  for i in [-1,0,1]: # loop over closest cells
    for j in [-1,0,1]: # loop over closest cells
      shifts.append(i*a1+j*a2) # list of possible vectors
      index.append((i,j)) # list of indexes
  def rot(r1,r2):
    dr = r1-r2
    rr = [dr + v for v in shifts] # loop over shifts
    rr2 = [r.dot(r) for r in rr] # distance
    mr = np.min(rr2) # minimum distance
    (i,j) = index[rr2.index(mr)] # index of the joining
    return v[0]*i + v[1]*j # return rotation
  return rot





def cubic_diamond_lattice():
  """Return a diamond lattice"""
  fcc = [] # fcc vectors
  fcc += [np.array([0.,0.,0.])]
  fcc += [np.array([0.5,0.5,0.])]
  fcc += [np.array([0.,0.5,0.5])]
  fcc += [np.array([0.5,0.,0.5])]
  rs = fcc + [r + np.array([.25,.25,.25]) for r in fcc] # all the positions
  fac = np.sqrt(3)/4. # distance to FN
  rs = [r/fac for r in rs] # positions
  g = Geometry() # create geometry
  g.a1 = np.array([1.,0.,0.])/fac # lattice vector
  g.a2 = np.array([0.,1.,0.])/fac # lattice vector
  g.a3 = np.array([0.,0.,1.])/fac # lattice vector
  g.dimensionality = 3 # three dimensional system
  g.has_sublattice = True
  g.sublattice = np.array([1 for i in range(4)] + [-1 for i in range(4)])
  g.r = np.array(rs) # store
  g.r2xyz() # create r coordinates
  g.get_fractional()
  return g



def diamond_lattice_minimal():
  """Return a diamond lattice"""
  fcc = [] # fcc vectors
  fcc += [np.array([0.,0.,0.])]
  rs = fcc + [r + np.array([-.25,.25,.25]) for r in fcc] # all the positions
  fac = np.sqrt(3)/4. # distance to FN
  rs = [r/fac for r in rs] # positions
  g = Geometry() # create geometry
  g.a1 = np.array([-.5,.5,0.])/fac # lattice vector
  g.a2 = np.array([0.,.5,.5])/fac # lattice vector
  g.a3 = np.array([-.5,0.,.5])/fac # lattice vector
  g.dimensionality = 3 # three dimensional system
  g.has_sublattice = True
  g.sublattice = np.array([1,-1])
  g.r = np.array(rs) # store
  g.r2xyz() # create r coordinates
  g.get_fractional()
  g = sculpt.set_xy_plane(g) # a1 and a2 in the xy plane
  return g




def pyrochlore_lattice():
  """Return a pyrochlore lattice"""
  rs = [np.array([0.,0.,0.])]
  rs += [np.array([-.25,.25,0.])]
  rs += [np.array([0.,.25,.25])]
  rs += [np.array([-.25,0.,.25])]
  fac = np.sqrt(rs[1].dot(rs[1])) # distance to FN
  rs = [np.array(r)/fac for r in rs] # positions
  g = Geometry() # create geometry
  g.a1 = np.array([-.5,.5,0.])/fac # lattice vector
  g.a2 = np.array([0.,.5,.5])/fac # lattice vector
  g.a3 = np.array([-.5,0.,.5])/fac # lattice vector
  g.dimensionality = 3 # three dimensional system
  g.has_sublattice = True
  g.sublattice_number = 4 # three sublattices
  g.sublattice = [1,0,3,2] # the three sublattices
#  g.sublattice = np.array([1,-1])
  g.r = np.array(rs) # store
  g.r2xyz() # create r coordinates
  g.center() # center the geometry
  g.get_fractional()
  return g



def tetrahedral_lattice():
  """Return a single layer of the pyrochlore lattice"""
  g = pyrochlore_lattice()
  from . import films
  g = films.geometry_film(g,nz=1)
  return g







# use the cubic one as the default one
#diamond_lattice = cubic_diamond_lattice 
diamond_lattice = diamond_lattice_minimal



# two dimensional geometries
geometries2d = [] 
geometries2d += [honeycomb_lattice]
geometries2d += [square_lattice]
geometries2d += [kagome_lattice]
geometries2d += [triangular_lattice]



from .geometrytk.sublattice import get_sublattice


from .neighbor import neighbor_directions
from .neighbor import neighbor_cells


def replicate_array(g,v,nrep=1):
   """Replicate a certain array in a supercell"""
   if len(np.array(v).shape)>1: # not one dimensional
       return np.array([replicate_array(g,vi,nrep=nrep) for vi in v.T]).T
   else: 
       from .checkclass import number2array
       nrep = number2array(nrep,d=g.dimensionality) # as array
       nout = 1
       for n in nrep: nout *= n # multiply
       return np.array(v.tolist()*nout)


def write_profile(g,d,name="PROFILE.OUT",nrep=3,normal_order=False):
  """Write a certain profile in a file"""
  if g.dimensionality == 0: nrep = 1
  if callable(d): d = np.array([d(ri) for ri in g.r]) # call
  else: d = np.array(d) # assume it is an array
  go = g.copy() # copy geometry
  go = go.supercell(nrep) # create supercell
  d = replicate_array(g,d,nrep=nrep) # replicate
  if normal_order:
      m = np.array([go.x,go.y,go.z,d]).T
      header = "x        y       z        profile"
  else:
      m = np.array([go.x,go.y,d,go.z]).T
      header = "x        y     profile      z"
  np.savetxt(name,m,fmt='%.8f',delimiter="    ",header=header) # save in file



from .indexing import get_index



def same_site(r1,r2):
    """Check if it is the same site"""
    dr = r1-r2
    dr = dr.dot(dr)
    if dr<0.0001: return 1.0
    else: return 0.0



def hyperhoneycomb_lattice():
  """Return a hyperhoneycomb lattice"""
  g = Geometry() # create geometry
  g.a1 = np.array([np.sqrt(3.),0.,0.]) # lattice vector
  g.a2 = np.array([0.,np.sqrt(3.),0.]) # lattice vector
  g.a3 = np.array([-np.sqrt(3.)/2.,np.sqrt(3.)/2.,3.]) # lattice vector
  rs = [] # vectors in the unit cell
  rs.append(np.array([0.,0.,-0.5])+g.a1/2.) # site
  rs.append(np.array([0.,0.,0.])) # site
  rs.append(np.array([0.,0.,1.])) # site
  rs.append(np.array([0.,0.,1.5])+g.a2/2.) # site
  g.dimensionality = 3 # three dimensional system
  g.has_sublattice = True
  g.sublattice = np.array([1,-1,1,-1])
  g.r = np.array(rs) # store
  g.r2xyz() # create r coordinates
  g.get_fractional()
  return g




from .geometrytk.write import write_vasp


from .neighbor import neighbor_distances


def array2function(g,v):
    r = g.r # positions
    def f(ri):
        return array2function_jit(r,v,np.array(ri))
    return f # return function


@jit(nopython=True)
def array2function_jit(r,v,ir):
    n = len(r)
    for i in range(n):
        dr = r[i] - ir # vector difference
        dr2 = dr[0]**2 + dr[1]**2 + dr[2]**2
        if dr2<1e-3: return v[i]
    return 0.0


from .sculpt import image2island


from .geometrytk import readgeometry 
read_xyz = readgeometry.read_xyz



def get_supercell(self,nsuper,store_primal=False):
    """Creates a supercell"""
    from .checkclass import number2array
    if store_primal: # store the primal geometry
        self.primal_geometry = self.copy() 
    if self.dimensionality==0: return self # zero dimensional
    if np.array(nsuper).shape==(3,3): # if a matrix is given
        return supercelltk.non_orthogonal_supercell(self,nsuper)
    if self.dimensionality==1:
        if checkclass.is_iterable(nsuper): nsuper = nsuper[0]
        return supercell1d(self,nsuper)
    elif self.dimensionality==2:
        nsuper = number2array(nsuper,d=2) # get an array
        nsuper1 = nsuper[0] 
        nsuper2 = nsuper[1]
        if np.max(np.abs(nsuper-np.round(nsuper)))>1e-5:
            return supercelltk.target_angle_volume(self,angle=None,
                    volume=nsuper1*nsuper2)
        else: return supercell2d(self,n1=nsuper1,n2=nsuper2)
    elif self.dimensionality==3:
        nsuper = number2array(nsuper,d=3)
        if np.max(np.abs(nsuper-np.round(nsuper)))>1e-5: raise # not implementet
        nsuper1 = nsuper[0]
        nsuper2 = nsuper[1]
        nsuper3 = nsuper[2]
        s = supercell3d(self,n1=nsuper1,n2=nsuper2,n3=nsuper3)
    else: raise # not implemented
    s.center()
    s.get_fractional()
    return s








from .kpointstk.mapping import get_k2K_generator


gdict = dict() # dictionary
gdict["chain"] = chain
gdict["square"] = square_lattice
gdict["honeycomb"] = honeycomb_lattice
gdict["triangular"] = triangular_lattice
gdict["kagome"] = kagome_lattice
gdict["lieb"] = lieb_lattice
gdict["pyrochlore"] = pyrochlore_lattice
gdict["diamond"] = diamond_lattice
gdict["cubic"] = cubic_lattice


def get_geometry(g):
    """Return a certain geometry"""
    if type(g)==Geometry: return g
    elif type(g)==str:
        if g in gdict: return gdict[g]() # return the geometry
    elif g is None: return get_geometry("square") # default geometry
    else: raise


def sierpinski(**kwargs):
    from .geometrytk import fractals
    return fractals.sierpinski(**kwargs)






# kernel polynomial method libraries
from __future__ import print_function,division
import scipy.sparse.linalg as lg
from scipy.sparse import csc_matrix as csc
import numpy.random as rand
from scipy.sparse import coo_matrix,csc_matrix,bmat
import numpy as np
from scipy.signal import hilbert
from . import algebra
from numba import jit

# check that the fortran library exists
try: 
  from . import kpmf90 
  use_fortran = True
except:
  use_fortran = False # use python routines




def get_moments(v,m,n=100,use_fortran=use_fortran,test=False):
    """ Get the first n moments of a certain vector
    using the Chebychev recursion relations"""
    if use_fortran:
        from .kpmf90 import get_momentsf90 # fortran routine
        mo = coo_matrix(m) # convert to coo matrix
        v = algebra.matrix2vector(v)
    # call the fortran routine
        mus = get_momentsf90(mo.row+1,mo.col+1,mo.data,v,n) 
        return mus # return fortran result
    else:
        if test: return python_kpm_moments_clear(v,m,n=n)
        else: return python_kpm_moments(v,m,n=n)


def python_kpm_moments(v,m,n=100):
    """Python routine to calculate moments"""
    mus = np.array([0.0j for i in range(2*n)]) # empty arrray for the moments
    am = v.copy() # zero vector
    a = m@v  # vector number 1
    bk = algebra.braket_ww(v,v)
  #  bk = (np.transpose(np.conjugate(v))*v)[0,0] # scalar product
    bk1 = algebra.braket_ww(a,v)
  #  bk1 = (np.transpose(np.conjugate(a))*v)[0,0] # scalar product
    
    mus[0] = bk.copy()  # mu0
    mus[1] = bk1.copy() # mu1
    for i in range(1,n): 
        ap = 2*m@a - am # recursion relation
        bk = algebra.braket_ww(a,a)
        bk1 = algebra.braket_ww(ap,a)
        mus[2*i] = 2.*bk
        mus[2*i+1] = 2.*bk1
        am = a.copy() # new variables
        a = ap.copy() # new variables
    mu0 = mus[0] # first
    mu1 = mus[1] # second
    for i in range(1,n): 
      mus[2*i] +=  - mu0
      mus[2*i+1] += -mu1 
    return mus


def python_kpm_moments_clear(v,m,n=100):
  """Python routine to calculate moments"""
  mus = np.array([0.0j for i in range(2*n)]) # empty arrray for the moments
  a0 = v.copy() # first vector
  am = v.copy() # first vector
  a = m*v  # vector number 1
  mus[0] = 1.  # mu0
  mu = (np.transpose(np.conjugate(a0))*a)[0,0] # scalar product
  mus[1] = mu # mu1
  for i in range(1,2*n): 
      ap = 2*m*a - am # recursion relation
      mu = (np.transpose(np.conjugate(a0))*a)[0,0] # scalar product
      mus[i] = mu # store
      am = a.copy() # new variables
      a = ap.copy() # new variables
  return mus






def get_momentsA(v,m,n=100,A=None):
    """ Get the first n moments of a certain vector
    using the Chebychev recursion relations"""
    if A is None: raise # only for a certain A
    mus = np.array([0.0j for i in range(n)]) # empty arrray for the moments
    v = algebra.matrix2vector(v) # zero vector
    A = csc_matrix(A) # turn sparse
    m = csc_matrix(m) # turn sparse
    return get_momentsA_jit(v,m,n,A,mus)

#@jit(nopython=True)
def get_momentsA_jit(v,m,n,A,mus):
    am = v*1.0
    a = m@v  # vector number 1
    bk = np.conjugate(v).dot(A@v) 
    bk1 = np.conjugate(a).dot(A@v)
    mus[0] = bk  # mu0
    mus[1] = bk1 # mu1
    for i in range(2,n): 
        ap = 2.*m@a - am # recursion relation
        bk = np.conjugate(ap).dot(A@v) 
        mus[i] = bk
        am = a.copy() # new variables
        a = ap.copy() # new variables
    mu0 = mus[0] # first
    mu1 = mus[1] # second
    return mus




def get_moments_ij(m0,n=100,i=0,j=0,use_fortran=use_fortran):
  """ Get the first n moments of a the |i><j| operator
  using the Chebychev recursion relations"""
  m = coo_matrix(m0,dtype=np.complex_)
  if use_fortran:
    mus = kpmf90.get_moments_ij(m.row+1,m.col+1,m.data,n,m.shape[0],i+1,j+1)
    return mus
  else:
    mus = np.zeros(n,dtype=np.complex_) # empty arrray for the moments
    v = np.zeros(m.shape[0],dtype=np.complex_) ; v[i] = 1.0 # initial vector
    v = np.matrix([v]).T # zero vector
    am = v.copy()
    a = m*v  # vector number 1
    bk = v[j] # scalar product
    bk1 = a[j,0] # scalar product
    mus[0] = bk  # mu0
    mus[1] = bk1 # mu1
    for ii in range(2,n): 
      ap = 2.*m*a - am # recursion relation
      bk = ap[j,0] # scalar product
      mus[ii] = bk
      am = a.copy() # new variables
      a = ap.copy() # new variables
    return mus

def get_moments_vivj(m0,vi,vj,n=100,use_fortran=False):
  if not use_fortran: return get_moments_vivj_python(m0,vi,vj,n=n)
  else: return get_moments_vivj_fortran(m0,vi,vj,n=n)


def get_moments_vivj_python(m0,vi,vj,n=100):
  """ Get the first n moments of a the |i><j| operator
  using the Chebychev recursion relations"""
  m = csc_matrix(m0,dtype=np.complex_)
  mus = np.zeros(n,dtype=np.complex_) # empty arrray for the moments
  v = vi.copy()
  am = v.copy()
  a = m@v  # vector number 1
  bk = algebra.braket_ww(vj,v)
#  bk = (vj.H*v).todense().trace()[0,0] # calculate bk
  bk1 = algebra.braket_ww(vj,a)
#  bk1 = (vj.H*a).todense().trace()[0,0] # calculate bk
  mus[0] = bk  # mu0
  mus[1] = bk1 # mu1
  for ii in range(2,n): 
    ap = 2.*m@a - am # recursion relation
    bk = algebra.braket_ww(vj,ap)
#    bk = (vj.H*ap).todense().trace()[0,0]
    mus[ii] = bk
    am = a.copy() # new variables
    a = ap.copy() # new variables
  return mus



def get_moments_vivj_fortran(m0,vi,vj,n=100):
    raise # I haven't check this function
    mo = coo_matrix(m0) # convert to coo matrix
    vi1 = vi.todense() # convert to conventional vector
    vj1 = vj.todense() # convert to conventional vector
# call the fortran routine
    mus = get_moments_vivj(mo.row+1,mo.col+1,mo.data,vi,vj,n) 
    return mus # return fortran result



def full_trace(m_in,n=200,use_fortran=use_fortran):
  """ Get full trace of the matrix"""
  m = csc(m_in) # saprse matrix
  nd = m.shape[0] # length of the matrix
  mus = np.array([0.0j for i in range(2*n)])
#  for i in range(ntries):
  for i in range(nd):
    mus += local_dos(m_in,i=i,n=n,use_fortran=use_fortran)
  return mus/nd









def local_dos(m_in,i=0,n=200,use_fortran=use_fortran):
  """ Calculates local DOS using the KPM"""
  m = csc(m_in) # saprse matrix
  nd = m.shape[0] # length of the matrix
  mus = np.array([0.0j for j in range(2*n)])
  v = rand.random(nd)*0.
  v[i] = 1.0 # vector only in site i 
  v = csc(v).transpose()
# get the chebychev moments
  mus += get_moments(v,m,n=n,use_fortran=use_fortran) 
  return mus



def ldos(m_in,i=0,scale=10.,npol=None,ne=500,kernel="jackson"):
  """Return two arrays with energies and local DOS"""
  if npol is None: npol = ne
  mus = local_dos(csc_matrix(m_in)/scale,i=i,n=npol) # get coefficients
  xs = np.linspace(-1.0,1.0,ne,endpoint=True)*0.99 # energies
  ys = generate_profile(mus,xs,kernel=kernel)
  return (scale*xs,ys/scale)



ldos0d = ldos



def tdos(m_in,scale=10.,npol=None,ne=500,kernel="jackson",
              ntries=20,ewindow=None,frand=None,
              operator=None,x=None):
  """Return two arrays with energies and local DOS"""
  if npol is None: npol = ne
  mus = random_trace(m_in/scale,ntries=ntries,n=npol,fun=frand,
          operator=operator) 
  if ewindow is None or abs(ewindow)>scale: # no window provided
    xs = np.linspace(-1.0,1.0,ne,endpoint=True)*1.01 # energies
  else:
    xx = abs(ewindow/scale) # scale
    xs = np.linspace(-xx,xx,ne,endpoint=True)*1.01 # energies
  ys = generate_profile(mus,xs,kernel=kernel).real
  (xs,ys) = (scale*xs,ys/scale)
  if x is not None:
    from scipy.interpolate import interp1d
    f = interp1d(xs,ys,bounds_error=False,fill_value=[ys[0],ys[-1]])
    return x,f(x)
  else: return xs,ys


def pdos(m,P=None,**kwargs):
    """Compute the projected density of states, assuming the operator
    fufills P^2 = P"""
    from .randomtk import randomwf
    fun0 = randomwf(m.shape[0]) # generator
    if P is not None: # operator provided
        from .operators import Operator
        op = Operator(P).get_matrix() # redefine
        fun = lambda : op@fun0() # define new generator
    else: fun = fun0 # original generator
    return tdos(m,frand=fun,**kwargs) # call TDOS with the generator



tdos0d = tdos # redefine


def total_energy(m_in,scale=10.,npol=None,ne=500,ntries=20):
   x,y = tdos0d(m_in,scale=scale,npol=npol,ne=ne,ntries=ntries)
   z = .5*(np.sign(x)+1.)*x*y # function to integrate
   return np.trapz(z,x)



def random_trace(m_in,ntries=20,n=200,fun=None,operator=None):
  """ Calculates local DOS using the KPM"""
  m = csc(m_in) # sparse matrix
  nd = m.shape[0] # length of the matrix
  if fun is not None: # check that dimensions are fine
    v0 = fun()
    if len(v0) != m_in.shape[0]: raise
  if fun is None:
#    def fun(): return rand.random(nd) -.5 + 1j*rand.random(nd) -.5j
      from .randomtk import randomwf
      fun = randomwf(nd) # generator
  def pfun(x):
    v = fun()
    v = v/np.sqrt(v.dot(np.conjugate(v))) # normalize the vector
#    v = csc(v).transpose()
    if operator is None:
        mus = get_moments(v,m,n=n) # get the chebychev moments
    else:
#        mus = get_moments_vivj(m,v,operator@v,n=2*n,use_fortran=False)
        mus = get_momentsA(v,m,n=2*n,A=operator) # get the chebychev moments
    return mus
  from . import parallel
  out = parallel.pcall(pfun,range(ntries))
  mus = np.zeros(out[0].shape,dtype=np.complex_)
  for o in out: mus = mus + o # add contribution
  return mus/ntries



def random_trace_A(m_in,ntries=20,n=200,A=None):
  """ Calculates local DOS using the KPM"""
  m = csc(m_in) # saprse matrix
  nd = m.shape[0] # length of the matrix
  mus = np.array([0.0j for j in range(n)])
  for i in range(ntries): # loop over tries
    #v = rand.random(nd) - .5
    v = rand.random(nd) -.5 + 1j*rand.random(nd) -.5j
    v = v/np.sqrt(v.dot(v)) # normalize the vector
    v = csc(v).transpose()
    mus += get_momentsA(v,m,n=n,A=A) # get the chebychev moments
  return mus/ntries



def full_trace_A(m_in,ntries=20,n=200,A=None):
  """ Calculates local DOS using the KPM"""
  m = csc(m_in) # saprse matrix
  nd = m.shape[0] # length of the matrix
  mus = np.array([0.0j for j in range(2*n)])
  for i in range(nd): # loop over tries
    #v = rand.random(nd) - .5
    v = rand.random(nd)*0.
    v[i] = 1.0 # vector only in site i 
    v = csc(v).transpose()
    mus += get_momentsA(v,m,n=n,A=A) # get the chebychev moments
  return mus/nd



def correlator0d(m_in,i=0,j=0,scale=10.,npol=None,ne=500,write=True,
  x=None):
  """Return two arrays with energies and local DOS"""
  if npol is None: npol = ne
  mus = get_moments_ij(m_in/scale,n=npol,i=i,j=j,use_fortran=True)
  if np.sum(np.abs(mus.imag))>0.001:
#    print("WARNING, off diagonal has nonzero imaginary elements",np.sum(np.abs(mus.imag)))
     pass
  if x is None: xs = np.linspace(-1.0,1.0,ne,endpoint=True)*0.99 # energies
  else: xs = x/scale # use from input
  ys = generate_green_profile(mus,xs,kernel="jackson",use_fortran=False)/scale*np.pi # so it is the Green function
#  imys = hilbert(ys).imag
  if write: np.savetxt("CORRELATOR_KPM.OUT",np.matrix([scale*xs,-ys.imag,ys.real]).T)
  return (scale*xs,ys.real,ys.imag)




def dm_ij_energy(m_in,i=0,j=0,scale=10.,npol=None,ne=500,x=None):
  """Return the correlation function"""
  if npol is None: npol = ne
  mus = get_moments_ij(m_in/scale,n=npol,i=i,j=j,use_fortran=use_fortran)
  if np.sum(np.abs(mus.imag))>0.001:
#    print("WARNING, off diagonal has nonzero imaginary elements",np.sum(np.abs(mus.imag)))
    pass
  if x is None: xs = np.linspace(-1.0,1.0,ne,endpoint=True)*0.99 # energies
  else: xs = x/scale # use from input
  ysr = generate_profile(mus.real,xs,kernel="jackson",use_fortran=use_fortran)/scale*np.pi # so it is the Green function
  ysi = generate_profile(mus.imag,xs,kernel="jackson",use_fortran=use_fortran)/scale*np.pi # so it is the Green function
  ys = ysr - 1j*ysi
  return (scale*xs,ys)



def dm_vivj_energy(m_in,vi,vj,scale=10.,npol=None,ne=500,x=None):
  """Return the correlation function"""
  if npol is None: npol = ne
  mus = get_moments_vivj(m_in/scale,vi,vj,n=npol)
  if np.sum(np.abs(mus.imag))>0.001:
#    print("WARNING, off diagonal has nonzero imaginary elements",np.sum(np.abs(mus.imag)))
    pass
  if x is None: xs = np.linspace(-1.0,1.0,ne,endpoint=True)*0.99 # energies
  else: xs = x/scale # use from input
  ysr = generate_profile(mus.real,xs,kernel="lorentz",use_fortran=use_fortran)/scale*np.pi # so it is the Green function
  ysi = generate_profile(mus.imag,xs,kernel="jackson",use_fortran=use_fortran)/scale*np.pi # so it is the Green function
  ys = ysr - 1j*ysi
  return (scale*xs,ys)







def generate_profile(mus,xs,kernel="jackson",use_fortran=use_fortran):
  """ Uses the Chebychev expansion to create a certain profile"""
  # initialize polynomials
#  xs = np.array([0.])
  tm = np.zeros(xs.shape) +1.
  t = xs.copy()
  if kernel=="jackson": mus = jackson_kernel(mus)
  elif kernel=="lorentz": mus = lorentz_kernel(mus)
  else: raise
  if use_fortran: # call the fortran routine
    ys = kpmf90.generate_profile(mus,xs) 
  else: # do a python loop
    ys = np.zeros(xs.shape,dtype=np.complex_) + mus[0] # first term
    # loop over all contributions
    for i in range(1,len(mus)):
      mu = mus[i]
      ys += 2.*mu*t # add contribution
      tp = 2.*xs*t - tm # chebychev recursion relation
      tm = t + 0.
      t = 0. + tp # next iteration
    ys = ys/np.sqrt(1.-xs*xs) # prefactor
  ys = ys/np.pi
  return ys




def generate_green_profile(mus,xs,kernel="jackson",use_fortran=use_fortran):
  """ Uses the Chebychev expansion to create a certain profile"""
  # initialize polynomials
#  xs = np.array([0.])
  tm = np.zeros(xs.shape) +1.
  t = xs.copy()
  ys = np.zeros(xs.shape,dtype=np.complex_) + mus[0]/2 # first term
  if kernel=="jackson": mus = jackson_kernel(mus)
  elif kernel=="lorentz": mus = lorentz_kernel(mus)
  else: raise
  if True:
    for i in range(1,len(mus)): # loop over mus
      ys += np.exp(1j*i*np.arccos(xs))*mus[i] # add contribution
    ys = ys/np.sqrt(1.-xs*xs)
    return 1j*2*ys/np.pi
#    if use_fortran: # call the fortran routine
#      ys = kpmf90.generate_profile(mus,xs) 
#    return ys








def dos(m_in,xs,ntries=20,n=200,scale=10.):
  """Return the density of states"""
  if scale is None: scale = 10.*np.max(np.abs(m_in.data)) # estimate of the value
  mus = random_trace(m_in/scale,ntries=ntries,n=n)
  ys = generate_profile(mus,xs/scale) # generate the DOS
  return ys # return the DOS 




from .kpmtk.kernels import fejer_kernel
from .kpmtk.kernels import lorentz_kernel
from .kpmtk.kernels import jackson_kernel



def edge_dos(intra0,inter0,scale=4.,w=20,npol=300,ne=500,bulk=False,
                use_random=True,nrand=20):
  """Calculated the edge DOS using the KPM"""
  h = [[None for j in range(w)] for i in range(w)]
  intra = csc_matrix(intra0)
  inter = csc_matrix(inter0)
  for i in range(w): h[i][i] = intra
  for i in range(w-1): 
    h[i+1][i] = inter.H
    h[i][i+1] = inter
  h = bmat(h) # sparse hamiltonian
  ds = np.zeros(ne)
  dsb = np.zeros(ne)
  norb = intra0.shape[0] # orbitals ina cell
  for i in range(norb):
    (xs,ys) = ldos0d(h,i=i,scale=scale,npol=npol,ne=ne) 
    ds += ys # store
    if bulk:
      (xs,zs) = ldos0d(h,i=w*norb//2 + i,scale=scale,npol=npol,ne=ne) 
      dsb += zs # store
  if not bulk: return (xs,ds/w)
  else: return (xs,ds/w,dsb/w)











# routines to align spins in the right direction
import numpy as np

tol = 1e-5 # tolerancy

def most_perpendicular_vector(vs):
    """Compute what is the most perpendicular vector to a set of vectors
    vs"""
    from scipy.optimize import minimize
    def fun(thetaphi):
        return perp_jax(thetaphi,vs)
    def grad_fun(thetaphi):
        return jac_jax(thetaphi,vs)
    x0 = np.random.random(2)
    res = minimize(fun,x0,tol=tol,jac=grad_fun)
    v = tp2v(res.x) # get the vector
    return v # return the vector


def most_perp_basis(vs):
    """Rotate the magnetization to the basis in which the magnetization
    is mostly contained in the xy plane"""
    v = most_perpendicular_vector(vs)
    out = random_perp_basis(v) # create change of basis
    ovs = np.array([out.T@v for v in vs])
#    print(v)
#    print(np.round(out,2))
#    print(ovs.shape)
    return np.array(ovs)




def random_perp_basis(v):
    """Generate a random perpendicular vector"""
    out = [v] # initial vector
    T = []
    for i in range(2):
        v0 = np.random.random(3) - 0.5 # random vector
        for vi in out:
            v0 = v0/np.sqrt(v0.dot(v0)) # normalize
            v0 = v0 - np.dot(v0,vi)*vi # remove component
        v0 = v0/np.sqrt(v0.dot(v0)) # normalize
#        print(i,v0)
        out.append(v0) # store
        T.append(v0)
    T.append(v)
    return np.array(T).T # new basis



def tp2v(thetaphi):
    """Convert Theta Phi to a vector"""
    theta = thetaphi[0]
    phi = thetaphi[1]
    vx = np.sin(theta)*np.cos(phi) # Mx
    vy = np.sin(theta)*np.sin(phi) # My
    vz = np.cos(theta) # Mz
    v0 = np.array([vx,vy,vz]) # full vector
    return v0
  
import jax.numpy as jnp
from jax import jit
from jax import grad

def perp_jax_master(thetaphi,vs):
    theta = thetaphi[0]
    phi = thetaphi[1]
    vx = jnp.sin(theta)*jnp.cos(phi) # Mx
    vy = jnp.sin(theta)*jnp.sin(phi) # My
    vz = jnp.cos(theta) # Mz
    v0 = jnp.array([vx,vy,vz]) # full vector
    n = len(vs) # loop over vectors
    out = 0. # storage
    for i in range(n): # loop over vectors
        v = vs[i,:] # get this vector
        w = jnp.cross(v,v0) # cross product
        out = out - jnp.sum(w*w) # compute the norm
    return out

perp_jax = jit(perp_jax_master) # jit jax function for energy
jac_jax = jit(grad(perp_jax_master,argnums=0)) # jit jax gradient



from __future__ import print_function
import numpy as np
from numba import jit

def bulk1d(g,fac=0.3):
  """Return the indexes of the bulk sites"""
  ymin = np.min(g.y)
  ymax = np.max(g.y)
  out = np.zeros(len(g.r)) # output array
  for i in range(len(g.r)):
    if fac<(g.y[i]-ymin)/(ymax-ymin)<(1.0-fac): out[i] = 1.0
  return out


def bulk0d(g0,fac=0.3):
  """Return the indexes of the bulk sites"""
  g = g0.copy()
  g.center()
  r2 = np.sqrt(g.x**2 + g.y**2 + g.z**2)
  rmax = np.max(r2)
  out = np.zeros(r2.shape) # initialize
  out[r2<(fac*rmax)] = 1.0
  return out


def bulk(g,**kwargs):
    if g.dimensionality==0:  return bulk0d(g,**kwargs)
    elif g.dimensionality==1:  return bulk1d(g,**kwargs)


def get_index(g,r,replicas=False):
    """Given a certain position, return the index of it in the geometry"""
    if replicas: # check the replicas
      ds = g.neighbor_directions()
      rset = [g.replicas(d) for d in ds]  # all sets of replicas
    else:
      rset = [g.r] # list of positions
    for rs in rset: # loop over set of sites
      out = get_index_jit(r,np.array(rs)) # index
      if out>0: 
          return out # valid index
    return None # not found


@jit(nopython=True)
def get_index_jit(r0,rs):
    for i in range(len(rs)): # loop
        dr = r0 - rs[i] # different
        dr2 = dr[0]*dr[0] + dr[1]*dr[1] + dr[2]*dr[2]
        if dr2<1e-5: return i
    return -1



# routines to reduce the size of the hamiltonian

def fullsc2minsc(hin):
  """ Reduces the size of the hamiltonian from full superconducting
  to a minimal """
  h = hin.copy() # copy hamiltonian
  n = len(h.intra)/4 # number of spinless orbitals
  m = h.intra
  def reduce_eh(m):
    t = np.matrix([[0j for i in range(2*n)] for j in range(2*n)])
    for i in range(n): # loop over spinless orbitals
      for j in range(n): # loop over spinless orbitals
        t[i,j] = m[2*i,2*j] # electron up
        t[i+n,j+n] = m[2*n+2*i+1,2*n+2*j+1]  # hole down
        t[i,j+n] = m[2*i+1,2*n+2*j+1]  # coupling 
        t[j+n,i] = np.conjugate(t[i,j+n])  # coupling 
    return t
  h.intra = reduce_eh(h.intra) # apply the operator
  if h.dimensionality==1:
    h.inter = reduce_eh(h.inter)
  if h.dimensionality==2:
    h.tx = reduce_eh(h.tx)
    h.ty = reduce_eh(h.ty)
    h.txy = reduce_eh(h.txy)
    h.txmy = reduce_eh(h.txmy)
  return h # return new hamiltonian 
  

